from .attribute import GetSetAttributeMixin
from .audio import AudioDataMixin
from .buffer import BufferDataMixin
from .content import ContentPropertyMixin
from .convert import ConvertMixin
from .dump import DumpFileMixin
from .image import ImageDataMixin
from .match import MatchMixin
from .mesh import MeshDataMixin
from .plot import PlotMixin
from .text import TextDataMixin
from .version import VersionedMixin
from .video import VideoDataMixin


class AllMixins(
    ContentPropertyMixin,
    ConvertMixin,
    AudioDataMixin,
    ImageDataMixin,
    TextDataMixin,
    MeshDataMixin,
    VideoDataMixin,
    BufferDataMixin,
    PlotMixin,
    DumpFileMixin,
    VersionedMixin,
    MatchMixin,
    GetSetAttributeMixin,
):
    """All plugins that can be used in :class:`Document`. """

    ...



import ast
import asyncio
import ipaddress
import socket
from abc import abstractmethod
from argparse import Namespace
from threading import Thread
from typing import Optional, TYPE_CHECKING

import grpc

from jina.logging.logger import JinaLogger
from jina.proto import jina_pb2_grpc
from jina.types.message import Message
from .. import __default_host__, __docker_host__
from ..helper import get_public_ip, get_internal_ip, get_or_reuse_loop

if TYPE_CHECKING:
    import kubernetes


class ConnectionList:
    """
    Maintains a list of connections and uses round roubin for selecting a connection

    :param port: port to use for the connections
    """

    def __init__(self, port: int):
        self.port = port
        self._connections = []
        self._address_to_connection_idx = {}
        self._rr_counter = 0

    def add_connection(self, address: str, connection):
        """
        Add connection with ip to the connection list
        :param address: Target address of this connection
        :param connection: The connection to add
        """
        if address not in self._address_to_connection_idx:
            self._address_to_connection_idx[address] = len(self._connections)
            self._connections.append(connection)

    def remove_connection(self, address: str):
        """
        Remove connection with ip from the connection list
        :param address: Remove connection for this address
        :returns: The removed connection or None if there was not any for the given ip
        """
        if address in self._address_to_connection_idx:
            self._rr_counter = (
                self._rr_counter % (len(self._connections) - 1)
                if (len(self._connections) - 1)
                else 0
            )
            return self._connections.pop(self._address_to_connection_idx.pop(address))

        return None

    def get_next_connection(self):
        """
        Returns a connection from the list. Strategy is round robin
        :returns: A connection from the pool
        """
        try:
            connection = self._connections[self._rr_counter]
        except IndexError:
            # This can happen as a race condition while removing connections
            self._rr_counter = 0
            connection = self._connections[self._rr_counter]
        self._rr_counter = (self._rr_counter + 1) % len(self._connections)
        return connection

    def pop_connection(self):
        """
        Removes and returns a connection from the list. Strategy is round robin
        :returns: The connection removed from the pool
        """
        if self._connections:
            connection = self._connections.pop(self._rr_counter)
            self._rr_counter = (
                (self._rr_counter + 1) % len(self._connections)
                if len(self._connections)
                else 0
            )
            return connection
        else:
            return None

    def has_connection(self, address: str) -> bool:
        """
        Checks if a connection for ip exists in the list
        :param address: The address to check
        :returns: True if a connection for the ip exists in the list
        """
        return address in self._address_to_connection_idx


class ConnectionPool:
    """
    Manages a list of connections.

    :param logger: the logger to use
    :param on_demand_connection: Flag to indicate if connections should be created on demand
    """

    def __init__(self, logger: Optional[JinaLogger] = None, on_demand_connection=True):
        self._connections = {}
        self._on_demand_connection = on_demand_connection

        self._logger = logger or JinaLogger(self.__class__.__name__)

    def send_message(self, msg: Message, target_address: str):
        """Send msg to target_address via one of the pooled connections

        :param msg: message to send
        :param target_address: address to send to, should include the port like 1.1.1.1:53
        :return: result of the actual send method
        """
        if target_address in self._connections:
            pooled_connection = self._connections[target_address].get_next_connection()
            return self._send_message(msg, pooled_connection)
        elif self._on_demand_connection:
            # If the pool is disabled and an unknown connection is requested: create it
            connection_pool = self._create_connection_pool(target_address)
            return self._send_message(msg, connection_pool.get_next_connection())
        else:
            raise ValueError(f'Unknown address {target_address}')

    def _create_connection_pool(self, target_address):
        port = target_address[target_address.rfind(':') + 1 :]
        connection_pool = ConnectionList(port=port)
        connection_pool.add_connection(
            target_address, self._create_connection(target=target_address)
        )
        self._connections[target_address] = connection_pool
        return connection_pool

    def start(self):
        """
        Starts the connection pool
        """
        pass

    def close(self):
        """
        Closes the connection pool
        """
        self._connections.clear()

    @abstractmethod
    def _send_message(self, msg: Message, connection):
        ...

    @abstractmethod
    def _create_connection(self, target):
        ...


class GrpcConnectionPool(ConnectionPool):
    """
    GrpcConnectionPool which uses gRPC as the communication mechanism
    """

    def _send_message(self, msg: Message, connection):
        # this wraps the awaitable object from grpc as a coroutine so it can be used as a task
        # the grpc call function is not a coroutine but some _AioCall
        async def task_wrapper(new_message, stub):
            await stub.Call(new_message)

        return asyncio.create_task(task_wrapper(msg, connection))

    def _create_connection(self, target):
        self._logger.debug(f'create connection to {target}')
        channel = grpc.aio.insecure_channel(
            target,
            options=[
                ('grpc.max_send_message_length', -1),
                ('grpc.max_receive_message_length', -1),
            ],
        )

        return jina_pb2_grpc.JinaDataRequestRPCStub(channel)


class K8sGrpcConnectionPool(GrpcConnectionPool):
    """
    Manages grpc connections to replicas in a K8s deployment.

    :param namespace: K8s namespace to operate in
    :param client: K8s client
    :param logger: the logger to use
    """

    def __init__(
        self,
        namespace: str,
        client: 'kubernetes.client.CoreV1Api',
        logger: JinaLogger = None,
    ):
        super().__init__(logger=logger, on_demand_connection=False)

        self._namespace = namespace
        self._deployment_clusteraddresses = {}
        self._k8s_client = client
        self._k8s_event_queue = asyncio.Queue()
        self.enabled = False

        self._fetch_initial_state()

        from kubernetes import watch

        self._api_watch = watch.Watch()

        self.update_thread = Thread(target=self.run, daemon=True)

    def _fetch_initial_state(self):
        namespaced_pods = self._k8s_client.list_namespaced_pod(self._namespace)
        for item in namespaced_pods.items:
            self._process_item(item)

    def start(self):
        """
        Subscribe to the K8s API and watch for changes in Pods
        """
        self._loop = get_or_reuse_loop()
        self._process_events_task = asyncio.create_task(self._process_events())
        self.update_thread.start()

    async def _process_events(self):
        while self.enabled:
            event = await self._k8s_event_queue.get()
            self._process_item(event)

    def run(self):
        """
        Subscribes on MODIFIED events from list_namespaced_pod AK8s PI
        """

        self.enabled = True
        while self.enabled:
            for event in self._api_watch.stream(
                self._k8s_client.list_namespaced_pod, self._namespace
            ):
                if event['type'] == 'MODIFIED':
                    asyncio.run_coroutine_threadsafe(
                        self._k8s_event_queue.put(event['object']), self._loop
                    )
                if not self.enabled:
                    break

    def close(self):
        """
        Closes the connection pool
        """
        self.enabled = False
        self._process_events_task.cancel()
        self._api_watch.stop()
        super().close()

    def send_message(self, msg: Message, target_address: str):
        """
        Send msg to target_address via one of the pooled connections.

        :param msg: message to send
        :param target_address: address to send to, should include the port like 1.1.1.1:53
        :return: result of the actual send method
        """
        host, port = target_address.split(':')
        # host can be a domain instead of IP Address, resolve it to IP Address
        return super().send_message(msg, f'{socket.gethostbyname(host)}:{port}')

    @staticmethod
    def _pod_is_up(item):
        return item.status.pod_ip is not None and item.status.phase == 'Running'

    @staticmethod
    def _pod_is_ready(item):
        return item.status.container_statuses is not None and all(
            cs.ready for cs in item.status.container_statuses
        )

    def _process_item(self, item):
        deployment_name = item.metadata.labels["app"]
        is_deleted = item.metadata.deletion_timestamp is not None

        if not is_deleted and self._pod_is_up(item) and self._pod_is_ready(item):
            if deployment_name in self._deployment_clusteraddresses:
                self._add_pod_connection(deployment_name, item)
            else:
                cluster_ip, port = self._find_cluster_ip(deployment_name)
                if cluster_ip:
                    self._deployment_clusteraddresses[
                        deployment_name
                    ] = f'{cluster_ip}:{port}'
                    self._connections[f'{cluster_ip}:{port}'] = ConnectionList(port)
                    self._add_pod_connection(deployment_name, item)
                else:
                    self._logger.debug(
                        f'Observed state change in unknown deployment {deployment_name}'
                    )
        elif (
            is_deleted
            and self._pod_is_up(item)
            and deployment_name in self._deployment_clusteraddresses
        ):
            self._remove_pod_connection(deployment_name, item)

    def _remove_pod_connection(self, deployment_name, item):
        target = item.status.pod_ip
        connection_pool = self._connections[
            self._deployment_clusteraddresses[deployment_name]
        ]
        if connection_pool.has_connection(f'{target}:{connection_pool.port}'):
            self._logger.debug(
                f'Removing connection to {target}:{connection_pool.port} for deployment {deployment_name} at {self._deployment_clusteraddresses[deployment_name]}'
            )
            self._connections[
                self._deployment_clusteraddresses[deployment_name]
            ].remove_connection(f'{target}:{connection_pool.port}')

    def _add_pod_connection(self, deployment_name, item):
        target = item.status.pod_ip
        connection_pool = self._connections[
            self._deployment_clusteraddresses[deployment_name]
        ]
        if not connection_pool.has_connection(f'{target}:{connection_pool.port}'):
            self._logger.debug(
                f'Adding connection to {target}:{connection_pool.port} for deployment {deployment_name} at {self._deployment_clusteraddresses[deployment_name]}'
            )

            connection_pool.add_connection(
                f'{target}:{connection_pool.port}',
                self._create_connection(target=f'{target}:{connection_pool.port}'),
            )

    def _extract_app(self, service_item):
        if service_item.metadata.annotations:
            return ast.literal_eval(
                list(service_item.metadata.annotations.values())[0]
            )['spec']['selector']['app']
        elif service_item.metadata.labels:
            return service_item.metadata.labels['app']

        return None

    def _find_cluster_ip(self, deployment_name):
        service_resp = self._k8s_client.list_namespaced_service(self._namespace)
        for s in service_resp.items:
            app = self._extract_app(s)
            if app and deployment_name == app and s.spec.cluster_ip:
                # find the port-in for this deployment
                for p in s.spec.ports:
                    if p.name == 'port-in':
                        return s.spec.cluster_ip, p.port

        return None, None


def is_remote_local_connection(first: str, second: str):
    """
    Decides, whether ``first`` is remote host and ``second`` is localhost

    :param first: the ip or host name of the first runtime
    :param second: the ip or host name of the second runtime
    :return: True, if first is remote and second is local
    """

    try:
        first_ip = ipaddress.ip_address(first)
        first_global = first_ip.is_global
    except ValueError:
        if first == 'localhost':
            first_global = False
        else:
            first_global = True
    try:
        second_ip = ipaddress.ip_address(second)
        second_local = second_ip.is_private or second_ip.is_loopback
    except ValueError:
        if second == 'localhost':
            second_local = True
        else:
            second_local = False

    return first_global and second_local


def get_connect_host(
    bind_host: str,
    bind_expose_public: bool,
    connect_args: Namespace,
) -> str:
    """
    Compute the host address for ``connect_args``

    :param bind_host: the ip for binding
    :param bind_expose_public: True, if bind socket should be exposed publicly
    :param connect_args: configuration for the host ip connection
    :return: host ip
    """
    runs_in_docker = connect_args.runs_in_docker
    # by default __default_host__ is 0.0.0.0

    # is BIND at local
    bind_local = bind_host == __default_host__

    # is CONNECT at local
    conn_local = connect_args.host == __default_host__

    # is CONNECT inside docker?
    # check if `uses` has 'docker://' or,
    # it is a remote pea managed by jinad. (all remote peas are inside docker)
    conn_docker = (
        (
            getattr(connect_args, 'uses', None) is not None
            and (
                connect_args.uses.startswith('docker://')
                or connect_args.uses.startswith('jinahub+docker://')
            )
        )
        or not conn_local
        or runs_in_docker
    )

    # is BIND & CONNECT all on the same remote?
    bind_conn_same_remote = (
        not bind_local and not conn_local and (bind_host == connect_args.host)
    )

    # pod1 in local, pod2 in local (conn_docker if pod2 in docker)
    if bind_local and conn_local:
        return __docker_host__ if conn_docker else __default_host__

    # pod1 and pod2 are remote but they are in the same host (pod2 is local w.r.t pod1)
    if bind_conn_same_remote:
        return __docker_host__ if conn_docker else __default_host__

    if bind_local and not conn_local:
        # in this case we are telling CONN (at remote) our local ip address
        if connect_args.host.startswith('localhost'):
            # this is for the "psuedo" remote tests to pass
            return __docker_host__
        return get_public_ip() if bind_expose_public else get_internal_ip()
    else:
        # in this case we (at local) need to know about remote the BIND address
        return bind_host


def create_connection_pool(args: 'Namespace') -> ConnectionPool:
    """
    Creates the appropriate connection pool based on args
    :param args: Arguments for this pod
    :return: A connection pool object
    """
    if args.k8s_namespace and args.k8s_connection_pool:
        from jina.peapods.pods.k8slib.kubernetes_client import K8sClients

        k8s_clients = K8sClients()
        return K8sGrpcConnectionPool(
            namespace=args.k8s_namespace,
            client=k8s_clients.core_v1,
        )
    else:
        return GrpcConnectionPool()



from contextlib import nullcontext
from typing import Union, BinaryIO, TYPE_CHECKING

from ..... import __windows__
from .....proto import jina_pb2

if TYPE_CHECKING:
    from ...document import DocumentArray
    from ...memmap import DocumentArrayMemmap


class BinaryIOMixin:
    """Save/load an array to a binary file. """

    @classmethod
    def load_binary(
        cls, file: Union[str, BinaryIO]
    ) -> Union['DocumentArray', 'DocumentArrayMemmap']:
        """Load array elements from a binary file.

        :param file: File or filename to which the data is saved.

        :return: a DocumentArray object
        """

        if hasattr(file, 'read'):
            file_ctx = nullcontext(file)
        else:
            file_ctx = open(file, 'rb')

        dap = jina_pb2.DocumentArrayProto()

        from ...document import DocumentArray

        with file_ctx as fp:
            dap.ParseFromString(fp.read())
            da = cls()
            da.extend(DocumentArray(dap.docs))
            return da

    def save_binary(self, file: Union[str, BinaryIO]) -> None:
        """Save array elements into a binary file.

        Comparing to :meth:`save_json`, it is faster and the file is smaller, but not human-readable.

        :param file: File or filename to which the data is saved.
        """
        if hasattr(file, 'write'):
            file_ctx = nullcontext(file)
        else:
            if __windows__:
                file_ctx = open(file, 'wb', newline='')
            else:
                file_ctx = open(file, 'wb')

        with file_ctx as fp:
            dap = jina_pb2.DocumentArrayProto()
            if self._pb_body:
                dap.docs.extend(self._pb_body)
            fp.write(dap.SerializePartialToString())



import warnings
from typing import List, Union, Any, Optional

from ....helper import T
from ....helper import dunder_get, typename
from ....proto import jina_pb2


class GetSetAttributeMixin:
    """Provide helper functions for :class:`Document` to allow advanced set and get attributes """

    _ON_GETATTR = ['matches', 'chunks']
    _special_mapped_keys = ('scores', 'evaluations')

    def __getattr__(self, item):
        if item in self._ON_GETATTR:
            self._increaseVersion()
        if hasattr(self._pb_body, item):
            value = getattr(self._pb_body, item)
        elif '__' in item:
            value = dunder_get(self._pb_body, item)
        else:
            raise AttributeError(f'no attribute named `{item}`')
        return value

    @classmethod
    def attributes(
        cls,
        include_proto_fields: bool = True,
        include_proto_fields_camelcase: bool = False,
        include_properties: bool = False,
    ) -> List[str]:
        """Return all attributes supported by the Document, which can be accessed by ``doc.attribute``

        :param include_proto_fields: if set, then include all protobuf fields
        :param include_proto_fields_camelcase: if set, then include all protobuf fields in CamelCase
        :param include_properties: if set, then include all properties defined for Document class
        :return: a list of attributes in string.
        """
        import inspect

        support_keys = []

        if include_proto_fields:
            support_keys = list(jina_pb2.DocumentProto().DESCRIPTOR.fields_by_name)
        if include_proto_fields_camelcase:
            support_keys += list(
                jina_pb2.DocumentProto().DESCRIPTOR.fields_by_camelcase_name
            )

        if include_properties:
            support_keys += [
                name
                for (name, value) in inspect.getmembers(
                    cls, lambda x: isinstance(x, property)
                )
            ]
        return list(set(support_keys))

    def get_attributes(self, *fields: str) -> Union[Any, List[Any]]:
        """Bulk fetch Document fields and return a list of the values of these fields

        .. note::
            Arguments will be extracted using `dunder_get`
            .. highlight:: python
            .. code-block:: python

                d = Document({'id': '123', 'hello': 'world', 'tags': {'id': 'external_id', 'good': 'bye'}})

                assert d.id == '123'  # true
                assert d.tags['hello'] == 'world' # true
                assert d.tags['good'] == 'bye' # true
                assert d.tags['id'] == 'external_id' # true

                res = d.get_attrs_values(*['id', 'tags__hello', 'tags__good', 'tags__id'])

                assert res == ['123', 'world', 'bye', 'external_id']

        :param fields: the variable length values to extract from the document
        :return: a list with the attributes of this document ordered as the args
        """

        ret = []
        for k in fields:
            try:
                value = getattr(self, k)

                if value is None:
                    raise ValueError

                ret.append(value)
            except (AttributeError, ValueError):
                warnings.warn(
                    f'Could not get attribute `{typename(self)}.{k}`, returning `None`'
                )
                ret.append(None)

        # unboxing if args is single
        if len(fields) == 1:
            ret = ret[0]

        return ret

    def set_attributes(self, **kwargs):
        """Bulk update Document fields with key-value specified in kwargs

        .. seealso::
            :meth:`get_attributes` for bulk get attributes

        :param kwargs: the keyword arguments to set the values, where the keys are the fields to set
        """
        for k, v in kwargs.items():
            if isinstance(v, (list, tuple)):
                if k == 'chunks':
                    self.chunks.extend(v)
                elif k == 'matches':
                    self.matches.extend(v)
                else:
                    self._pb_body.ClearField(k)
                    getattr(self._pb_body, k).extend(v)
            elif isinstance(v, dict) and k not in self._special_mapped_keys:
                self._pb_body.ClearField(k)
                getattr(self._pb_body, k).update(v)
            else:
                cls = type(self)
                if (
                    hasattr(cls, k)
                    and isinstance(getattr(cls, k), property)
                    and getattr(cls, k).fset
                ):
                    # if class property has a setter
                    setattr(self, k, v)
                elif hasattr(self._pb_body, k):
                    # no property setter, but proto has this attribute so fallback to proto
                    setattr(self._pb_body, k, v)
                else:
                    raise AttributeError(f'{k} is not recognized')

    def update(
        self: T,
        source: T,
        fields: Optional[List[str]] = None,
    ) -> None:
        """Updates fields specified in ``fields`` from the source to current Document.

        :param source: The :class:`Document` we want to update from as source. The current
            :class:`Document` is referred as destination.
        :param fields: a list of field names that we want to update, if not specified,
            use all present fields in source.

        .. note::
            *. if ``fields`` are empty, then all present fields in source will be merged into current document.
            * `tags` will be updated like a python :attr:`dict`.
            *. the current :class:`Document` will be modified in place, ``source`` will be unchanged.
            *. if current document has more fields than :attr:`source`, these extra fields wll be preserved.
        """
        # We do a safe update: only update existent (value being set) fields from source.
        present_fields = [
            field_descriptor.name
            for field_descriptor, _ in source._pb_body.ListFields()
        ]
        if not fields:
            fields = present_fields  # if `fields` empty, update all present fields.
        for field in fields:
            if (
                field == 'tags'
            ):  # For the tags, stay consistent with the python update method.
                self.tags.update(source.tags)
            else:
                self._pb_body.ClearField(field)
                try:
                    setattr(self, field, getattr(source, field))
                except AttributeError:
                    setattr(self._pb_body, field, getattr(source, field))



from typing import overload, TYPE_CHECKING, Union, Callable, Optional, Tuple

if TYPE_CHECKING:
    from ...arrays import DocumentArray
    from ...arrays.memmap import DocumentArrayMemmap
    from ...ndarray import ArrayType
    import numpy as np


class MatchMixin:
    """Provide sugary syntax for :class:`Document` to match against a :class:`DocumentArray`"""

    @overload
    def match(
        self,
        darray: Union['DocumentArray', 'DocumentArrayMemmap'],
        metric: Union[
            str, Callable[['ArrayType', 'ArrayType'], 'np.ndarray']
        ] = 'cosine',
        limit: Optional[Union[int, float]] = 20,
        normalization: Optional[Tuple[float, float]] = None,
        metric_name: Optional[str] = None,
        batch_size: Optional[int] = None,
        exclude_self: bool = False,
        only_id: bool = False,
        use_scipy: bool = False,
    ):
        """Matching the current Document against a set of Documents.

        The result will be stored in :attr:`.matches`.

        .. note::
            When you want to match a set Documents (let's call it set `A`) against another set of Documents (set `B`),
            where you want to find for each element in `A` what are its nearest neighbours in `B`.
            Then you need :meth:`DocumentArray.match`

        :param darray: the other DocumentArray or DocumentArrayMemmap to match against
        :param metric: the distance metric
        :param limit: the maximum number of matches, when not given defaults to 20.
        :param normalization: a tuple [a, b] to be used with min-max normalization,
                                the min distance will be rescaled to `a`, the max distance will be rescaled to `b`
                                all values will be rescaled into range `[a, b]`.
        :param metric_name: if provided, then match result will be marked with this string.
        :param batch_size: if provided, then `darray` is loaded in chunks of, at most, batch_size elements. This option
                           will be slower but more memory efficient. Specialy indicated if `darray` is a big
                           DocumentArrayMemmap.
        :param exclude_self: if set, Documents in ``darray`` with same ``id`` as the left-hand values will not be
                        considered as matches.
        :param only_id: if set, then returning matches will only contain ``id``
        :param use_scipy: if set, use ``scipy`` as the computation backend. Note, ``scipy`` does not support distance
            on sparse matrix.
        """
        ...

    def match(self, *args, **kwargs):
        """
        # noqa: D102
        # noqa: DAR101
        """
        from ...arrays import DocumentArray

        _tmp = DocumentArray([self])
        _tmp.match(*args, **kwargs)



from math import sqrt, ceil, floor
from typing import Optional

import numpy as np

from .... import Document
from ....helper import deprecated_method
from ....logging.profile import ProgressBar


class PlotMixin:
    """Helper functions for plotting the arrays. """

    @deprecated_method(new_function_name='plot_embeddings')
    def visualize(self, *args, **kwargs):
        """Deprecated! Please use :meth:`.plot_embeddings` instead.

        Plot embeddings in a 2D projection with the PCA algorithm. This function requires ``matplotlib`` installed.

        :param args: extra args
        :param kwargs: extra kwargs
        """
        self.plot_embeddings(*args, **kwargs)

    def plot_embeddings(
        self,
        output: Optional[str] = None,
        title: Optional[str] = None,
        colored_attr: Optional[str] = None,
        colormap: str = 'rainbow',
        method: str = 'pca',
        show_axis: bool = False,
        **kwargs,
    ):
        """Plot embeddings in a 2D projection with the PCA algorithm. This function requires ``matplotlib`` installed.

        If `tag_name` is provided the plot uses a distinct color for each unique tag value in the
        documents of the DocumentArray.

        :param output: Optional path to store the visualization. If not given, show in UI
        :param title: Optional title of the plot. When not given, the default title is used.
        :param colored_attr: Optional str that specifies attribute used to color the plot, it supports dunder expression
            such as `tags__label`, `matches__0__id`.
        :param colormap: the colormap string supported by matplotlib.
        :param method: the visualization method, available `pca`, `tsne`. `pca` is fast but may not well represent
                nonlinear relationship of high-dimensional data. `tsne` requires scikit-learn to be installed and is
                much slower.
        :param show_axis: If set, axis and bounding box of the plot will be printed.
        :param kwargs: extra kwargs pass to matplotlib.plot
        """

        x_mat = self.embeddings
        assert isinstance(
            x_mat, np.ndarray
        ), f'Type {type(x_mat)} not currently supported, use np.ndarray embeddings'

        if method == 'tsne':
            from sklearn.manifold import TSNE

            x_mat_2d = TSNE(n_components=2).fit_transform(x_mat)
        else:
            from ....math.dimensionality_reduction import PCA

            x_mat_2d = PCA(n_components=2).fit_transform(x_mat)

        plt_kwargs = {
            'x': x_mat_2d[:, 0],
            'y': x_mat_2d[:, 1],
            'alpha': 0.2,
            'marker': '.',
        }

        import matplotlib.pyplot as plt

        plt.figure(figsize=(8, 8))

        plt.title(title or f'{len(x_mat)} Documents with {method}')

        if colored_attr:
            tags = []

            for x in self:
                try:
                    tags.append(getattr(x, colored_attr))
                except (KeyError, AttributeError, ValueError):
                    tags.append(None)
            tag_to_num = {tag: num for num, tag in enumerate(set(tags))}
            plt_kwargs['c'] = np.array([tag_to_num[ni] for ni in tags])
            plt_kwargs['cmap'] = plt.get_cmap(colormap)

        # update the plt_kwargs
        plt_kwargs.update(kwargs)

        plt.scatter(**plt_kwargs)

        if not show_axis:
            plt.gca().set_axis_off()
            plt.gca().xaxis.set_major_locator(plt.NullLocator())
            plt.gca().yaxis.set_major_locator(plt.NullLocator())

        if output:
            plt.savefig(output, bbox_inches='tight', pad_inches=0.1)
        else:
            plt.show()

    def plot_image_sprites(
        self,
        output: Optional[str] = None,
        canvas_size: int = 512,
        min_size: int = 16,
        channel_axis: int = -1,
    ):
        """Generate a sprite image for all image blobs in this DocumentArray-like object.

        An image sprite is a collection of images put into a single image. It is always square-sized.
        Each sub-image is also square-sized and equally-sized.

        :param output: Optional path to store the visualization. If not given, show in UI
        :param canvas_size: the size of the canvas
        :param min_size: the minimum size of the image
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        """
        if not self:
            raise ValueError(f'{self!r} is empty')

        import matplotlib.pyplot as plt

        img_per_row = ceil(sqrt(len(self)))
        img_size = int(canvas_size / img_per_row)

        if img_size < min_size:
            # image is too small, recompute the size
            img_size = min_size
            img_per_row = int(canvas_size / img_size)

        max_num_img = img_per_row ** 2
        sprite_img = np.zeros(
            [img_size * img_per_row, img_size * img_per_row, 3], dtype='uint8'
        )
        img_id = 0

        actual_num_img = min(len(self), max_num_img)

        with ProgressBar(
            description='Generating sprite', total_length=actual_num_img
        ) as pg:
            for d in self:
                _d = Document(d, copy=True)
                if _d.content_type != 'blob':
                    _d.convert_uri_to_image_blob()
                    channel_axis = -1

                _d.set_image_blob_channel_axis(channel_axis, -1).set_image_blob_shape(
                    shape=(img_size, img_size)
                )

                row_id = floor(img_id / img_per_row)
                col_id = img_id % img_per_row
                sprite_img[
                    (row_id * img_size) : ((row_id + 1) * img_size),
                    (col_id * img_size) : ((col_id + 1) * img_size),
                ] = _d.blob

                img_id += 1
                pg.update()
                if img_id >= max_num_img:
                    break

        plt.gca().set_axis_off()
        plt.subplots_adjust(top=1, bottom=0, right=1, left=0, hspace=0, wspace=0)
        plt.margins(0, 0)
        plt.gca().xaxis.set_major_locator(plt.NullLocator())
        plt.gca().yaxis.set_major_locator(plt.NullLocator())

        plt.imshow(sprite_img)
        if output:
            plt.savefig(output, bbox_inches='tight', pad_inches=0.1, transparent=True)
        else:
            plt.show()



from typing import (
    TypeVar,
    Sequence,
)

from ..document import Document
from ...proto import jina_pb2

DocumentArraySourceType = TypeVar(
    'DocumentArraySourceType',
    jina_pb2.DocumentArrayProto,
    Sequence[Document],
    Sequence[jina_pb2.DocumentProto],
    Document,
)



import os
import re
import socket
import platform
from typing import Dict, List, Tuple, TYPE_CHECKING, Optional

import docker

from jina import __docker_host__
from jina.helper import colored
from jina.logging.logger import JinaLogger
from jina.peapods.runtimes.container.helper import get_gpu_device_requests
from . import (
    jinad_args,
    __root_workspace__,
    __partial_workspace__,
)
from .excepts import (
    DockerNotFoundException,
    DockerImageException,
    DockerNetworkException,
    DockerContainerException,
)
from .helper import id_cleaner, classproperty, is_error_message
from .models import DaemonID
from .models.enums import IDLiterals

if TYPE_CHECKING:
    from .files import DaemonFile
    from docker.models.networks import Network
    from docker.models.containers import Container
    from docker.client import APIClient, DockerClient


PORT_REGEX = r'[0-9]+(?:\.[0-9]+){3}:[0-9]+'


class Dockerizer:
    """Helper class to interact with docker client & dockerd"""

    logger = JinaLogger('Dockerizer', **vars(jinad_args))
    try:
        client: 'DockerClient' = docker.from_env()
        raw_client: 'APIClient' = docker.APIClient(
            base_url='unix://var/run/docker.sock'
        )
    except docker.errors.DockerException:
        logger.critical(
            f'docker client cannot connect to dockerd. '
            f'please start jinad with `-v /var/run/docker.sock:/var/run/docker.sock`'
        )
        raise DockerNotFoundException()

    @classmethod
    def _daemonize(cls, ids: List, attrs: str):
        for id in ids:
            try:
                field = getattr(id, attrs)
                if isinstance(field, List):
                    for f in field:
                        f = f.replace(':', '-')
                        yield DaemonID(f)
                else:
                    yield DaemonID(field)
            except TypeError:
                continue

    @classproperty
    def networks(cls) -> List[DaemonID]:
        """
        Returns all local docker networks with name in a pattern of `DaemonID`
        :return: list of networks created by jinad
        """
        return list(cls._daemonize(cls.client.networks.list(), 'name'))

    @classproperty
    def images(cls) -> List[DaemonID]:
        """
        Returns all local docker images with name in a pattern of `DaemonID`
        :return: list of images created by jinad
        """
        return list(cls._daemonize(cls.client.images.list(), 'tags'))

    @classproperty
    def containers(cls) -> List[DaemonID]:
        """
        Returns all local docker containers with name in a pattern of `DaemonID`
        :return: list of containers created by jinad
        """
        return list(cls._daemonize(cls.client.containers.list(), 'name'))

    @classmethod
    def network(cls, workspace_id: DaemonID) -> str:
        """
        Create a docker bridge network with name `workspace_id` using a predefined ipam config.
        All containers under `workspace_id` would use this network.
        :param workspace_id: workspace id
        :raises DockerNetworkException: if there are issues during network creation
        :return: id of the network
        """
        if workspace_id in cls.networks:
            network = cls.client.networks.get(network_id=workspace_id)
        else:
            from .stores import workspace_store

            new_subnet_start = (
                workspace_store.status.ip_range_start
                + workspace_store.status.ip_range_current_offset
            )

            ipam_config = docker.types.IPAMConfig(
                pool_configs=[
                    docker.types.IPAMPool(
                        subnet=f'{new_subnet_start}/{workspace_store.status.subnet_size}',
                        gateway=f'{new_subnet_start + 1}',
                    )
                ]
            )
            workspace_store.status.ip_range_current_offset += 2 ** (
                32 - workspace_store.status.subnet_size
            )
            try:
                network: 'Network' = cls.client.networks.create(
                    name=workspace_id, driver='bridge', ipam=ipam_config
                )
            except docker.errors.APIError as e:
                import traceback

                traceback.print_exc()
                cls.logger.critical(f'{e!r} during docker network creation')
                raise DockerNetworkException()
        return network.id

    @classmethod
    def build(
        cls, workspace_id: 'DaemonID', daemon_file: 'DaemonFile', logger: 'JinaLogger'
    ) -> str:
        """
        Build docker image using daemon file & tag it with `workspace_id`
        :param workspace_id: workspace id
        :param daemon_file: daemon file describing content inside the workdir
        :param logger: logger to be used
        :raises DockerImageException: if there are issues during image building
        :return: id of the image
        """
        logger.info(f'about to build image using {daemon_file}')

        def _log_stream(chunk, key):
            # logging taken from hubio.build
            _stream = chunk[key].splitlines()[0]
            if _stream:
                if is_error_message(_stream):
                    logger.critical(_stream)
                elif 'warning' in _stream.lower():
                    logger.warning(_stream)
                else:
                    logger.info(_stream)

        for build_log in cls.raw_client.build(
            path=daemon_file.dockercontext,
            dockerfile=daemon_file.dockerfile,
            buildargs=daemon_file.dockerargs,
            tag=workspace_id.tag,
            rm=True,
            pull=False,
            decode=True,
        ):
            if 'stream' in build_log:
                _log_stream(build_log, 'stream')
            elif 'message' in build_log:
                _log_stream(build_log, 'message')
            # elif 'status' in build_log:
            #     _log_stream(build_log, 'status')

        try:
            image = cls.client.images.get(name=workspace_id.tag)
        except docker.errors.ImageNotFound as e:
            logger.critical(f'Couldn\'t find image with name: {workspace_id.tag} {e!r}')
            raise DockerImageException(e)
        return id_cleaner(image.id)

    @classmethod
    def run_custom(
        cls, workspace_id: DaemonID, daemon_file: 'DaemonFile'
    ) -> Tuple['Container', str, Dict]:
        """Run a custom container during workspace creation.
        .. note::
            This invalidates the default entrypint (mini-jinad) & uses the entrypoint provided
            mentioned in the .jinad file (`run` section)
        :param workspace_id: workspace id
        :param daemon_file: daemon file describing content inside the workdir
        :return: tuple of container object, network id & ports
        """
        return cls.run(
            workspace_id=workspace_id,
            container_id=workspace_id,
            ports={f'{port}/tcp': port for port in daemon_file.ports},
            entrypoint=daemon_file.run,
        )

    @classmethod
    def run(
        cls,
        workspace_id: DaemonID,
        container_id: DaemonID,
        entrypoint: str,
        ports: Dict,
        envs: Optional[Dict] = {},
        device_requests: Optional[List] = None,
    ) -> Tuple['Container', str, Dict]:
        """
        Runs a container using an existing image (tagged with `workspace_id`).
        Maps `ports` to local dockerhost & tags the container with name `container_id`
        .. note::
            This uses the default entrypoint (mini-jinad) & appends `command` for execution.
        :param workspace_id: workspace id
        :param container_id: name of the container
        :param entrypoint: entrypoint for the container
        :param ports: ports to be mapped with local
        :param envs: dict of env vars to be set in the container
        :param device_requests: docker device requests
        :raises DockerImageException: if image is not found locally
        :raises DockerContainerException: if container creation fails
        :return: tuple of container object, network id & ports
        """

        def _validate_port_conflict(error: str) -> str:
            msg = ""
            if 'port is already allocated' in error:
                match = re.findall(PORT_REGEX, error)
                if match and len(match) > 0:
                    msg = f'port conflict: {match[0]}'
            return msg

        def _validate_device_request(error: str) -> bool:
            return True if 'could not select device driver' in error else False

        from .stores import workspace_store

        metadata = workspace_store[workspace_id].metadata
        if not metadata:
            raise DockerImageException(
                'Docker image not built properly, cannot proceed for run'
            )
        image = cls.client.images.get(name=metadata.image_id)
        network = metadata.network
        cls.logger.info(
            f'creating a container using image {colored(metadata.image_id, "cyan")} in network '
            f'{colored(network, "cyan")} and ports {colored(ports, "cyan")}'
        )
        try:
            run_kwargs = dict(
                image=image,
                name=container_id,
                volumes=cls.volume(workspace_id),
                environment=cls.environment(envs),
                network=network,
                ports=ports,
                detach=True,
                entrypoint=entrypoint,
                device_requests=device_requests or [],
                working_dir=__partial_workspace__,
                extra_hosts={__docker_host__: 'host-gateway'},
            )
            container: 'Container' = cls.client.containers.run(**run_kwargs)
        except docker.errors.NotFound as e:
            cls.logger.critical(
                f'Image {image} or Network {network} not found locally {e!r}'
            )

            raise DockerImageException(
                'Docker image not built properly, cannot proceed for run'
            )
        except docker.errors.APIError as e:
            msg = f'API Error while starting the docker container {e}'
            if _validate_device_request(str(e)):
                # It might not be possible for local to know remote machine's gpu status.
                # For few cases (Flow creation), we always set `gpus='all'`, which might fail
                # if dockerd cannot find the device. In that case, run again without device requests
                cls.logger.debug(
                    f'couldn\'t start the container with following device request. '
                    f'restarting after resetting device: {run_kwargs["device_requests"]}'
                )
                try:
                    # This leaves the old container in a "created" state and conflicts the name.
                    cls.rm_container(container_id)
                except:
                    pass
                run_kwargs.pop('device_requests')
                container: 'Container' = cls.client.containers.run(**run_kwargs)
            else:
                msg += ' ' + _validate_port_conflict(str(e))
                raise DockerContainerException(msg)
        except docker.errors.APIError as e:
            msg = f'API Error while starting the docker container{e}'
            msg += _validate_port_conflict(str(e))
            raise DockerContainerException(msg)
        # TODO: network & ports return can be avoided?
        return container, network, ports

    @classmethod
    def logs(cls, id: str) -> str:
        """Get all logs of a container

        :param id: container id
        :return: logs as str
        """
        try:
            container: 'Container' = cls.client.containers.get(container_id=id)
            return container.logs(stdout=True, stderr=True).decode()
        except docker.errors.NotFound:
            cls.logger.error(f'no containers with id {id} found')
            return ""

    @classmethod
    def _get_volume_host_dir(cls):
        try:
            volumes = cls.client.containers.get(socket.gethostname()).attrs[
                'HostConfig'
            ]['Binds']
            for volume in volumes:
                if volume.split(':')[1] == __root_workspace__:
                    return volume.split(':')[0]
        except:
            # above logic only works inside docker, if it does not work we dont need it
            pass
        return __root_workspace__

    @classproperty
    def dockersock(cls) -> str:
        """docker socket path

        :return: abs path to docker socket
        """
        location = '/var/run/docker.sock'
        return location if platform.system() == 'Darwin' else '/' + location

    @classmethod
    def volume(cls, workspace_id: DaemonID) -> Dict[str, Dict]:
        """
        Local volumes to be mounted inside the container during `run`.
        .. note::
            Local workspace should always be mounted to default WORKDIR for the container (/workspace).
            docker sock on dockerhost should also be mounted to make sure DIND works
        :param workspace_id: workspace id
        :return: dict of volume mappings
        """
        return {
            f'{cls._get_volume_host_dir()}/{workspace_id}': {
                'bind': __partial_workspace__,
                'mode': 'rw',
            },
            cls.dockersock: {'bind': '/var/run/docker.sock'},
        }

    @classmethod
    def environment(cls, envs: Dict[str, str]) -> Dict[str, str]:
        """
        Environment variables to be set inside the container during `run`

        :param envs: dict of env vars to be set in the container
        :return: dict of env vars
        """
        return {
            'JINA_LOG_WORKSPACE': os.path.join(__partial_workspace__, 'logs'),
            'JINA_LOG_LEVEL': os.getenv('JINA_LOG_LEVEL') or 'INFO',
            'JINA_HUB_ROOT': os.path.join(
                __partial_workspace__, '.jina', 'hub-packages'
            ),
            'JINA_HUB_CACHE_DIR': os.path.join(__partial_workspace__, '.cache', 'jina'),
            'HOME': __partial_workspace__,
            **envs,
        }

    @classmethod
    def remove(cls, id: DaemonID):
        """
        Determines type of jinad object & removes that from dockerd
        :param id: `DaemonID` describing local docker object
        """
        if id.jtype == IDLiterals.JNETWORK:
            cls.rm_network(id)
        elif id.jtype == IDLiterals.JWORKSPACE:
            cls.rm_image(id)
        else:
            cls.rm_container(id)

    @classmethod
    def containers_in_network(cls, id: str) -> List:
        """Get all containers currently connected to network
        :param id: network id
        :return: list of containers connected to network id
        """
        return [
            container["Name"]
            for container in cls.raw_client.inspect_network(net_id=id)[
                'Containers'
            ].values()
        ]

    @classmethod
    def rm_network(cls, id: str) -> bool:
        """
        Remove network from local if no containers are connected
        :param id: network id
        :return: True if deletion is successful else False
        """
        try:
            containers = cls.containers_in_network(id)
            if containers:
                cls.logger.info(
                    f'following containers are still connected to the network. skipping delete {containers}'
                )
                return False

            network: 'Network' = cls.client.networks.get(id)
            network.remove()
            cls.logger.success(f'network {colored(id, "cyan")} successfully removed')
            return True
        except docker.errors.NotFound as e:
            cls.logger.error(f'Couldn\'t find a network with id: `{id}`')
            return False
        except docker.errors.APIError as e:
            cls.logger.warning(
                f'dockerd threw an error while removing the network. '
                f'There might be containers still connected to the network `{id}`: \n{e}'
            )
            return False

    @classmethod
    def rm_image(cls, id: str):
        """
        Remove image from local
        :param id: image id
        :raises KeyError: if image is not found on local
        :raises DockerImageException: error during image removal
        """
        try:
            # TODO: decide when to force
            cls.client.images.remove(id, force=True)
        except docker.errors.ImageNotFound as e:
            cls.logger.error(f'Couldn\'t fetch image with name: `{id}`')
            raise KeyError(f'image `{id}` not found')
        except docker.errors.APIError as e:
            cls.logger.error(
                f'dockerd threw an error while removing the image `{id}`: {e}'
            )
            raise DockerImageException(f'dockerd error while removing image {id} {e!r}')

    @classmethod
    def rm_container(cls, id: str):
        """
        Remove container from local
        :param id: container id
        :raises KeyError: if container is not found on local
        :raises DockerContainerException: error during container removal
        """
        try:
            container: 'Container' = cls.client.containers.get(id)
            container.stop()
            container.remove()
        except docker.errors.NotFound as e:
            cls.logger.error(f'Couldn\'t fetch container with name: `{id}`')
            raise KeyError(f'container `{id}` not found')
        except docker.errors.APIError as e:
            cls.logger.error(
                f'dockerd threw an error while removing the container `{id}`: {e}'
            )
            raise DockerContainerException(
                f'dockerd error while removing network {id} {e!r}'
            )

    @classmethod
    def _validate(cls):
        # TODO
        pass



import base64
import io
import struct
from typing import Optional, Tuple, Union, BinaryIO

import numpy as np

from .helper import _get_file_context, _uri_to_buffer, _deprecate_by
from ....helper import T


class ImageDataMixin:
    """Provide helper functions for :class:`Document` to support image data. """

    def set_image_blob_channel_axis(
        self: T, original_channel_axis: int, new_channel_axis: int
    ) -> T:
        """Move the channel axis of the image :attr:`.blob` inplace.

        :param original_channel_axis: the original axis of the channel
        :param new_channel_axis: the new axis of the channel

        :return: itself after processed
        """
        self.blob = _move_channel_axis(
            self.blob, original_channel_axis, new_channel_axis
        )
        return self

    def convert_buffer_to_image_blob(
        self: T,
        width: Optional[int] = None,
        height: Optional[int] = None,
        channel_axis: int = -1,
    ) -> T:
        """Convert an image :attr:`.buffer` to a ndarray :attr:`.blob`.

        :param width: the width of the image blob.
        :param height: the height of the blob.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis

        :return: itself after processed
        """
        blob = _to_image_blob(io.BytesIO(self.buffer), width=width, height=height)
        blob = _move_channel_axis(blob, original_channel_axis=channel_axis)
        self.blob = blob
        return self

    def convert_image_blob_to_uri(self: T) -> T:
        """Assuming :attr:`.blob` is a _valid_ image, set :attr:`uri` accordingly

        :return: itself after processed
        """
        png_bytes = _to_png_buffer(self.blob)
        self.uri = 'data:image/png;base64,' + base64.b64encode(png_bytes).decode()
        return self

    def set_image_blob_shape(
        self: T,
        shape: Tuple[int, int],
        channel_axis: int = -1,
    ) -> T:
        """Resample the image :attr:`.blob` into different size inplace.

        If your current image blob has shape ``[H,W,C]``, then the new blob will be ``[*shape, C]``

        :param shape: the new shape of the image blob.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis

        :return: itself after processed
        """
        blob = _move_channel_axis(self.blob, original_channel_axis=channel_axis)
        out_rows, out_cols = shape
        in_rows, in_cols, n_in = blob.shape

        # compute coordinates to resample
        x = np.tile(np.linspace(0, in_cols - 2, out_cols), out_rows)
        y = np.repeat(np.linspace(0, in_rows - 2, out_rows), out_cols)

        # resample each image
        r = _nn_interpolate_2D(blob, x, y)
        self.blob = r.reshape(out_rows, out_cols, n_in)

        return self

    def dump_image_blob_to_file(
        self: T,
        file: Union[str, BinaryIO],
        channel_axis: int = -1,
    ) -> T:
        """Save :attr:`.blob` into a file

        :param file: File or filename to which the data is saved.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis

        :return: itself after processed
        """
        fp = _get_file_context(file)
        with fp:
            blob = _move_channel_axis(self.blob, channel_axis, -1)
            buffer = _to_png_buffer(blob)
            fp.write(buffer)
        return self

    def load_uri_to_image_blob(
        self: T,
        width: Optional[int] = None,
        height: Optional[int] = None,
        channel_axis: int = -1,
    ) -> T:
        """Convert the image-like :attr:`.uri` into :attr:`.blob`

        :param width: the width of the image blob.
        :param height: the height of the blob.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis

        :return: itself after processed
        """

        buffer = _uri_to_buffer(self.uri)
        blob = _to_image_blob(io.BytesIO(buffer), width=width, height=height)
        self.blob = _move_channel_axis(blob, original_channel_axis=channel_axis)
        return self

    def set_image_blob_normalization(
        self: T,
        channel_axis: int = -1,
        img_mean: Tuple[float] = (0.485, 0.456, 0.406),
        img_std: Tuple[float] = (0.229, 0.224, 0.225),
    ) -> T:
        """Normalize a uint8 image :attr:`.blob` into a float32 image :attr:`.blob` inplace.

        Following Pytorch standard, the image must be in the shape of shape (3 x H x W) and
        will be normalized in to a range of [0, 1] and then
        normalized using mean = [0.485, 0.456, 0.406] and std = [0.229, 0.224, 0.225]. These two arrays are computed
        based on millions of images. If you want to train from scratch on your own dataset, you can calculate the new
        mean and std. Otherwise, using the Imagenet pretrianed model with its own mean and std is recommended.

        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis
        :param img_mean: the mean of all images
        :param img_std: the standard deviation of all images
        :return: itself after processed

        .. warning::
            Please do NOT generalize this function to gray scale, black/white image, it does not make any sense for
            non RGB image. if you look at their MNIST examples, the mean and stddev are 1-dimensional
            (since the inputs are greyscale-- no RGB channels).


        """
        if self.blob.dtype == np.uint8 and self.blob.ndim == 3:
            blob = (self.blob / 255.0).astype(np.float32)
            blob = _move_channel_axis(blob, channel_axis, 0)
            mean = np.asarray(img_mean, dtype=np.float32)
            std = np.asarray(img_std, dtype=np.float32)
            blob = (blob - mean[:, None, None]) / std[:, None, None]
            # set back channel to original
            blob = _move_channel_axis(blob, 0, channel_axis)
            self.blob = blob
        else:
            raise ValueError(
                f'`blob` must be a uint8 ndarray with ndim=3, but receiving {self.blob.dtype} with ndim={self.blob.ndim}'
            )
        return self

    def convert_image_blob_to_sliding_windows(
        self: T,
        window_shape: Tuple[int, int] = (64, 64),
        strides: Optional[Tuple[int, int]] = None,
        padding: bool = False,
        channel_axis: int = -1,
        as_chunks: bool = False,
    ) -> T:
        """Convert :attr:`.blob` into a sliding window view with the given window shape :attr:`.blob` inplace.

        :param window_shape: desired output size. If size is a sequence like (h, w), the output size will be matched to
            this. If size is an int, the output will have the same height and width as the `target_size`.
        :param strides: the strides between two neighboring sliding windows. `strides` is a sequence like (h, w), in
            which denote the strides on the vertical and the horizontal axis. When not given, using `window_shape`
        :param padding: If False, only patches which are fully contained in the input image are included. If True,
            all patches whose starting point is inside the input are included, and areas outside the input default to
            zero. The `padding` argument has no effect on the size of each patch, it determines how many patches are
            extracted. Default is False.
        :param channel_axis: the axis id of the color channel, ``-1`` indicates the color channel info at the last axis.
        :param as_chunks: If set, each sliding window will be stored in the chunk of the current Document
        :return: Document itself after processed
        """
        window_h, window_w = window_shape
        stride_h, stride_w = strides or window_shape
        blob = _move_channel_axis(self.blob, channel_axis, -1)
        if padding:
            h, w, c = blob.shape
            ext_h = window_h - h % stride_h
            ext_w = window_w - w % window_w
            blob = np.pad(
                blob,
                ((0, ext_h), (0, ext_w), (0, 0)),
                mode='constant',
                constant_values=0,
            )
        h, w, c = blob.shape
        row_step = blob.strides[0]
        col_step = blob.strides[1]

        expanded_img = np.lib.stride_tricks.as_strided(
            blob,
            shape=(
                1 + int((h - window_h) / stride_h),
                1 + int((w - window_w) / stride_w),
                window_h,
                window_w,
                c,
            ),
            strides=(row_step * stride_h, col_step * stride_w, row_step, col_step, 1),
            writeable=False,
        )
        cur_loc_h, cur_loc_w = 0, 0
        if self.location:
            cur_loc_h, cur_loc_w = self.location[:2]

        bbox_locations = [
            (h * stride_h + cur_loc_h, w * stride_w + cur_loc_w, window_h, window_w)
            for h in range(expanded_img.shape[0])
            for w in range(expanded_img.shape[1])
        ]
        expanded_img = expanded_img.reshape((-1, window_h, window_w, c))
        if as_chunks:
            from . import Document

            for location, _blob in zip(bbox_locations, expanded_img):
                self.chunks.append(
                    Document(
                        blob=_move_channel_axis(_blob, -1, channel_axis),
                        location=location,
                    )
                )
        else:
            self.blob = _move_channel_axis(expanded_img, -1, channel_axis)
        return self

    convert_uri_to_image_blob = _deprecate_by(load_uri_to_image_blob)  #: Deprecated!


def _move_channel_axis(
    blob: np.ndarray, original_channel_axis: int = -1, target_channel_axis: int = -1
) -> np.ndarray:
    """This will always make the channel axis to the last of the :attr:`.blob`

    #noqa: DAR101
    #noqa: DAR201
    """
    if original_channel_axis != target_channel_axis:
        blob = np.moveaxis(blob, original_channel_axis, target_channel_axis)
    return blob


def _to_image_blob(
    source,
    width: Optional[int] = None,
    height: Optional[int] = None,
) -> 'np.ndarray':
    """
    Convert an image buffer to blob

    :param source: binary buffer or file path
    :param width: the width of the image blob.
    :param height: the height of the blob.
    :return: image blob
    """
    from PIL import Image

    raw_img = Image.open(source)
    if width or height:
        new_width = width or raw_img.width
        new_height = height or raw_img.height
        raw_img = raw_img.resize((new_width, new_height))
    try:
        return np.array(raw_img.convert('RGB'))
    except:
        return np.array(raw_img)


def _to_png_buffer(arr: 'np.ndarray') -> bytes:
    """
    Convert png to buffer bytes.

    :param arr: Data representations of the png.
    :return: Png in buffer bytes.

    ..note::
        if both :attr:`width` and :attr:`height` were provided, will not resize. Otherwise, will get image size
        by :attr:`arr` shape and apply resize method :attr:`resize_method`.
    """
    arr = arr.astype(np.uint8).squeeze()

    if arr.ndim == 1:
        # note this should be only used for MNIST/FashionMNIST dataset, because of the nature of these two datasets
        # no other image data should flattened into 1-dim array.
        png_bytes = _png_to_buffer_1d(arr, 28, 28)
    elif arr.ndim == 2:
        from PIL import Image

        im = Image.fromarray(arr).convert('L')
        png_bytes = _pillow_image_to_buffer(im, image_format='PNG')
    elif arr.ndim == 3:
        from PIL import Image

        im = Image.fromarray(arr).convert('RGB')
        png_bytes = _pillow_image_to_buffer(im, image_format='PNG')
    else:
        raise ValueError(
            f'{arr.shape} ndarray can not be converted into an image buffer.'
        )

    return png_bytes


def _png_to_buffer_1d(arr: 'np.ndarray', width: int, height: int) -> bytes:
    import zlib

    pixels = []
    for p in arr[::-1]:
        pixels.extend([p, p, p, 255])
    buf = bytearray(pixels)

    # reverse the vertical line order and add null bytes at the start
    width_byte_4 = width * 4
    raw_data = b''.join(
        b'\x00' + buf[span : span + width_byte_4]
        for span in range((height - 1) * width_byte_4, -1, -width_byte_4)
    )

    def png_pack(png_tag, data):
        chunk_head = png_tag + data
        return (
            struct.pack('!I', len(data))
            + chunk_head
            + struct.pack('!I', 0xFFFFFFFF & zlib.crc32(chunk_head))
        )

    png_bytes = b''.join(
        [
            b'\x89PNG\r\n\x1a\n',
            png_pack(b'IHDR', struct.pack('!2I5B', width, height, 8, 6, 0, 0, 0)),
            png_pack(b'IDAT', zlib.compress(raw_data, 9)),
            png_pack(b'IEND', b''),
        ]
    )

    return png_bytes


def _pillow_image_to_buffer(image, image_format: str) -> bytes:
    img_byte_arr = io.BytesIO()
    image.save(img_byte_arr, format=image_format)
    img_byte_arr = img_byte_arr.getvalue()
    return img_byte_arr


def _nn_interpolate_2D(X, x, y):
    nx, ny = np.around(x), np.around(y)
    nx = np.clip(nx, 0, X.shape[1] - 1).astype(int)
    ny = np.clip(ny, 0, X.shape[0] - 1).astype(int)
    return X[ny, nx, :]



import random
from collections import defaultdict
from typing import Dict, Any, TYPE_CHECKING, Generator

import numpy as np

if TYPE_CHECKING:
    from ..document import DocumentArray


class GroupMixin:
    """These helpers yield groups of :class:`DocumentArray` from
    a source :class:`DocumentArray` or :class:`DocumentArrayMemmap`."""

    def split(self, tag: str) -> Dict[Any, 'DocumentArray']:
        """Split the `DocumentArray` into multiple DocumentArray according to the tag value of each `Document`.

        :param tag: the tag name to split stored in tags.
        :return: a dict where Documents with the same value on `tag` are grouped together, their orders
            are preserved from the original :class:`DocumentArray`.

        .. note::
            If the :attr:`tags` of :class:`Document` do not contains the specified :attr:`tag`,
            return an empty dict.
        """
        from ..document import DocumentArray
        from ....helper import dunder_get

        rv = defaultdict(DocumentArray)
        for doc in self:
            if '__' in tag:
                value = dunder_get(doc.tags, tag)
            else:
                value = doc.tags.get(tag, None)

            if value is None:
                continue
            rv[value].append(doc)
        return dict(rv)

    def batch(
        self,
        batch_size: int,
        shuffle: bool = False,
    ) -> Generator['DocumentArray', None, None]:
        """
        Creates a `Generator` that yields `DocumentArray` of size `batch_size` until `docs` is fully traversed along
        the `traversal_path`. The None `docs` are filtered out and optionally the `docs` can be filtered by checking for
        the existence of a `Document` attribute.
        Note, that the last batch might be smaller than `batch_size`.

        :param batch_size: Size of each generated batch (except the last one, which might be smaller, default: 32)
        :param shuffle: If set, shuffle the Documents before dividing into minibatches.
        :yield: a Generator of `DocumentArray`, each in the length of `batch_size`
        """

        if not (isinstance(batch_size, int) and batch_size > 0):
            raise ValueError('`batch_size` should be a positive integer')

        N = len(self)
        ix = list(range(N))
        n_batches = int(np.ceil(N / batch_size))

        if shuffle:
            random.shuffle(ix)

        for i in range(n_batches):
            yield self[ix[i * batch_size : (i + 1) * batch_size]]



from typing import Iterable, TYPE_CHECKING

from .document import DocumentArray

if TYPE_CHECKING:
    from ..document import Document


class ChunkArray(DocumentArray):
    """
    :class:`ChunkArray` inherits from :class:`DocumentArray`.
    It's a subset of Documents.

    :param doc_views: Set of sub-documents (i.e chunks) of `reference_doc`
    :param reference_doc: Reference :class:`Document` for the sub-documents
    """

    def __init__(self, doc_views, reference_doc: 'Document'):
        """
        Set constructor method.

        :param doc_views: protobuf representation of the chunks
        :param reference_doc: parent document
        """
        self._ref_doc = reference_doc
        super().__init__(doc_views)

    def append(self, document: 'Document', **kwargs) -> 'Document':
        """Add a sub-document (i.e chunk) to the current Document.

        :param document: Sub-document to be appended
        :param kwargs: additional keyword arguments
        :return: the newly added sub-document in :class:`Document` view
        :rtype: :class:`Document`

        .. note::
            Comparing to :attr:`DocumentArray.append()`, this method adds more safeguard to
            make sure the added chunk is legit.
        """
        super().append(document)
        chunk = self[-1]
        chunk.set_attributes(
            parent_id=self._ref_doc.id, granularity=self.granularity, **kwargs
        )

        return chunk

    def extend(self, iterable: Iterable['Document']) -> None:
        """
        Extend the :class:`DocumentArray` by appending all the items from the iterable.

        :param iterable: the iterable of Documents to extend this array with
        """
        for doc in iterable:
            self.append(doc)

    @property
    def reference_doc(self) -> 'Document':
        """
        Get the document that :class:`ChunkArray` belongs to.

        :return: reference doc
        """
        return self._ref_doc

    @property
    def granularity(self) -> int:
        """
        Get granularity of all document in this array.

        :return: granularity
        """
        return self._ref_doc.granularity + 1

    @property
    def adjacency(self) -> int:
        """
        Get adjacency of all document in this array.

        :return: adjacency
        """
        return self._ref_doc.adjacency



import asyncio
import argparse
from abc import ABC, abstractmethod
from typing import (
    List,
    Dict,
    Union,
    Awaitable,
    Iterator,
    AsyncIterator,
    TYPE_CHECKING,
)

from .helper import AsyncRequestsIterator
from ...helper import get_or_reuse_loop
from ...logging.logger import JinaLogger
from ...types.message import Message

__all__ = ['BaseStreamer']

if TYPE_CHECKING:
    from ..grpc import Grpclet
    from ..zmq import AsyncZmqlet
    from ...types.request import Request, Response
    from ...clients.base.helper import HTTPClientlet, WebsocketClientlet


class BaseStreamer(ABC):
    """An base async request/response handler"""

    def __init__(
        self,
        args: argparse.Namespace,
        iolet: Union['AsyncZmqlet', 'Grpclet', 'HTTPClientlet', 'WebsocketClientlet'],
    ):
        """
        :param args: args from CLI
        :param iolet: One of AsyncZmqlet or Grpclet. Used for sending/receiving data to/from the Flow
        """
        self.args = args
        self.iolet = iolet
        self.logger = JinaLogger(self.__class__.__name__, **vars(args))
        self.request_buffer: Dict[str, asyncio.Future] = dict()
        self.receive_task = get_or_reuse_loop().create_task(self._receive())

    @abstractmethod
    async def _receive(self) -> Awaitable:
        """Receive background task"""
        ...

    @abstractmethod
    def _convert_to_message(self, request: 'Request') -> Union['Message', 'Request']:
        """Convert request to iolet message

        :param request: current request in the iterator
        """
        ...

    def _handle_request(self, request: 'Request') -> 'asyncio.Future':
        """
        For zmq & grpc data requests from gateway, for each request in the iterator, we send the `Message`
        using `iolet.send_message()`.

        For websocket requests from client, for each request in the iterator, we send the request in `bytes`
        using using `iolet.send_message()`.

        Then add {<request-id>: <an-empty-future>} to the request buffer.
        This empty future is used to track the `result` of this request during `receive`.

        :param request: current request in the iterator
        :return: asyncio Future for sending message
        """
        future = get_or_reuse_loop().create_future()
        self.request_buffer[request.request_id] = future
        asyncio.create_task(self.iolet.send_message(self._convert_to_message(request)))
        return future

    def _handle_response(self, response: 'Response') -> None:
        """Set result of each response received in the request buffer

        :param response: response received during `iolet.recv_message`
        """
        if response.request_id in self.request_buffer:
            future = self.request_buffer.pop(response.request_id)
            future.set_result(response)
        else:
            self.logger.warning(
                f'discarding unexpected response with request id {response.request_id}'
            )

    def _handle_end_of_iter(self) -> None:
        """Send end of iterator signal to Gateway"""
        pass

    @abstractmethod
    async def stream(
        self, request_iterator: Union[Iterator, AsyncIterator]
    ) -> AsyncIterator:
        """iterate through the request iterator and return responses in an async iterator

        :param request_iterator: requests iterator from Client
        """
        ...

    async def _stream_requests(
        self, request_iterator: Union[Iterator, AsyncIterator]
    ) -> AsyncIterator:
        """Implements request and response handling without prefetching

        :param request_iterator: requests iterator from Client
        :yield: responses
        """
        result_queue = asyncio.Queue()
        end_of_iter = asyncio.Event()
        futures = []

        def callback(future: 'asyncio.Future'):
            """callback to be run after future is completed.
            1. Put the future in the result queue.
            2. Remove the future from futures when future is completed.

            ..note::
                callback cannot be an awaitable, hence we cannot do `await queue.put(...)` here.
                We don't add `future.result()` to the queue, as that would consume the exception in the callback,
                which is difficult to handle.

            :param future: asyncio Future object retured from `handle_response`
            """
            result_queue.put_nowait(future)
            futures.remove(future)

        async def iterate_requests() -> None:
            """
            1. Traverse through the request iterator.
            2. `add_done_callback` to the future returned by `handle_request`.
                This callback adds the completed future to `result_queue`
            3. Append future to list of futures.
            4. Handle EOI (needed for websocket client)
            5. Set `end_of_iter` event
            """
            async for request in AsyncRequestsIterator(iterator=request_iterator):
                future: 'asyncio.Future' = self._handle_request(request=request)
                future.add_done_callback(callback)
                futures.append(future)
            self._handle_end_of_iter()
            end_of_iter.set()

        asyncio.create_task(iterate_requests())
        while not end_of_iter.is_set() or len(futures) > 0 or not result_queue.empty():
            # `not end_of_iter.is_set()` validates iterator is completed.
            # `len(futures) > 0` makes sure all futures are taken care of.
            # `not result_queue.empty()` makes sure all items in queue are processed.
            try:
                response: 'asyncio.Future' = result_queue.get_nowait()
                result_queue.task_done()
                yield response.result()
            except asyncio.QueueEmpty:
                await asyncio.sleep(0.2)
                continue

    async def _stream_requests_with_prefetch(
        self, request_iterator: Union[Iterator, AsyncIterator], prefetch: int
    ):
        """Implements request and response handling without prefetching

        :param request_iterator: requests iterator from Client
        :param prefetch: number of requests to prefetch
        :yield: response
        """

        async def iterate_requests(
            num_req: int, fetch_to: List[Union['asyncio.Task', 'asyncio.Future']]
        ):
            """
            1. Traverse through the request iterator.
            2. Append the future returned from `handle_request` to `fetch_to` which will later be awaited.

            :param num_req: number of requests
            :param fetch_to: the task list storing requests
            :return: False if append task to `fetch_to` else False
            """
            count = 0
            async for request in AsyncRequestsIterator(iterator=request_iterator):
                fetch_to.append(self._handle_request(request=request))
                count += 1
                if count == num_req:
                    return False
            return True

        prefetch_task = []
        is_req_empty = await iterate_requests(prefetch, prefetch_task)
        if is_req_empty and not prefetch_task:
            self.logger.error(
                'receive an empty stream from the client! '
                'please check your client\'s inputs, '
                'you can use "Client.check_input(inputs)"'
            )
            return

        # the total num requests < prefetch
        if is_req_empty:
            for r in asyncio.as_completed(prefetch_task):
                yield await r
        else:
            # if there are left over (`else` clause above is unnecessary for code but for better readability)
            onrecv_task = []
            # the following code "interleaves" prefetch_task and onrecv_task, when one dries, it switches to the other
            while prefetch_task:
                if self.logger.debug_enabled:
                    self.logger.debug(
                        f'send: {self.iolet.msg_sent} '
                        f'recv: {self.iolet.msg_recv} '
                        f'pending: {self.iolet.msg_sent - self.iolet.msg_recv}'
                    )
                onrecv_task.clear()
                for r in asyncio.as_completed(prefetch_task):
                    yield await r
                    if not is_req_empty:
                        is_req_empty = await iterate_requests(1, onrecv_task)

                # this list dries, clear it and feed it with on_recv_task
                prefetch_task.clear()
                prefetch_task = [j for j in onrecv_task]



from typing import Iterable, TYPE_CHECKING

__all__ = ['DocGroundtruthPair']

if TYPE_CHECKING:
    from . import Document


class DocGroundtruthPair:
    """
    Helper class to expose common interface to the traversal logic of the BaseExecutable Driver.
    It is important to note that it checks the matching structure of `docs` and `groundtruths`. It is important while
    traversing to ensure that then the driver can be applied at a comparable level of granularity and adjacency.
    This does not imply that you can't compare at the end a document with 10 matches with a groundtruth with 20 matches

    :param doc: Target `Document`.
    :param groundtruth: The :class:`Document` with desired state.
    """

    def __init__(self, doc: 'Document', groundtruth: 'Document'):
        """Set constructor method.

        :param doc: actual Document
        :param groundtruth: groundtruth Document
        """
        self.doc = doc
        self.groundtruth = groundtruth

    @property
    def matches(self) -> Iterable['DocGroundtruthPair']:
        """Get the pairs between matches and Groundtruth.

        :yields: DocGroundtruthPair object
        """
        assert len(self.doc.matches) == len(self.groundtruth.matches)
        for doc, groundtruth in zip(self.doc.matches, self.groundtruth.matches):
            yield DocGroundtruthPair(doc, groundtruth)

    @property
    def chunks(self) -> Iterable['DocGroundtruthPair']:
        """Get the pairs between chunks and Groundtruth.

        :yields: DocGroundtruthPair object
        """
        assert len(self.doc.chunks) == len(self.groundtruth.chunks)
        for doc, groundtruth in zip(self.doc.chunks, self.groundtruth.chunks):
            yield DocGroundtruthPair(doc, groundtruth)



import argparse

from typing import TYPE_CHECKING, AsyncIterator

from .base import BaseStreamer
from ...types.message import Message

__all__ = ['ZmqGatewayStreamer', 'GrpcGatewayStreamer']

if TYPE_CHECKING:
    from ..grpc import Grpclet
    from ...types.request import Request, Response


class GatewayStreamer(BaseStreamer):
    """Streamer used at Gateway to stream requests/responses to/from Executors"""

    def _convert_to_message(self, request: 'Request') -> Message:
        """Convert `Request` to `Message`

        :param request: current request in the iterator
        :return: Message object
        """
        return Message(None, request, 'gateway', **vars(self.args))

    async def stream(self, request_iterator, *args) -> AsyncIterator['Request']:
        """
        stream requests from client iterator and stream responses back.

        :param request_iterator: iterator of requests
        :param args: positional arguments
        :yield: responses from Executors
        """
        if self.receive_task.done():
            raise RuntimeError('receive task not running, can not send messages')

        async_iter: AsyncIterator = (
            self._stream_requests_with_prefetch(request_iterator, self.args.prefetch)
            if self.args.prefetch > 0
            else self._stream_requests(request_iterator)
        )

        async for response in async_iter:
            yield response

    # alias of stream used as a grpc servicer
    Call = stream


class ZmqGatewayStreamer(GatewayStreamer):
    """Streamer used at Gateway to stream ZMQ requests/responses to/from Executors"""

    async def _receive(self):
        """Await messages back from Executors and process them in the request buffer"""
        try:
            while True:
                response = await self.iolet.recv_message(callback=lambda x: x.response)
                # during shutdown the socket will return None
                if response is None:
                    break

                self._handle_response(response)
        finally:
            if self.request_buffer:
                self.logger.warning(
                    f'{self.__class__.__name__} closed, cancelling all outstanding requests'
                )
                for future in self.request_buffer.values():
                    future.cancel()
                self.request_buffer.clear()


class GrpcGatewayStreamer(GatewayStreamer):
    """Streamer used at Gateway to stream GRPC requests/responses to/from Executors"""

    def __init__(self, args: argparse.Namespace, iolet: 'Grpclet'):
        super().__init__(args, iolet)
        self.iolet.callback = lambda response: self._handle_response(response.request)

    async def _receive(self):
        """Start grpclet and await termination

        :return: await iolet start
        """
        return await self.iolet.start()

    async def _handle_response(self, response: 'Response'):
        """
        Async version of parents handle_response function

        :param response: message received from grpclet callback
        """
        super()._handle_response(response)



class K8sClients:
    """
    The Kubernetes api is wrapped into a class to have a lazy reading of the cluster configuration.

    """

    def __init__(self):
        import kubernetes

        try:
            # try loading kube config from disk first
            kubernetes.config.load_kube_config()
        except kubernetes.config.config_exception.ConfigException:
            # if the config could not be read from disk, try loading in cluster config
            # this works if we are running inside k8s
            kubernetes.config.load_incluster_config()

        self._k8s_client = kubernetes.client.ApiClient()
        self._core_v1 = None
        self._apps_v1 = None
        self._beta = None
        self._networking_v1_beta1_api = None

    @property
    def k8s_client(self):
        """Client for making requests to Kubernetes

        :return: k8s client
        """
        return self._k8s_client

    @property
    def core_v1(self):
        """V1 client for core

        :return: v1 client
        """
        if not self._core_v1:
            from kubernetes import client

            self._core_v1 = client.CoreV1Api(api_client=self._k8s_client)
        return self._core_v1

    @property
    def apps_v1(self):
        """V1 client for core

        :return: v1 client
        """
        if not self._apps_v1:
            from kubernetes import client

            self._apps_v1 = client.AppsV1Api(api_client=self._k8s_client)
        return self._apps_v1

    @property
    def beta(self):
        """Beta client for using beta features

        :return: beta client
        """
        if not self._beta:
            from kubernetes import client

            self._beta = client.ExtensionsV1beta1Api(api_client=self._k8s_client)
        return self._beta

    @property
    def networking_v1_beta1_api(self):
        """Networking client used for creating the ingress

        :return: networking client
        """
        if not self._networking_v1_beta1_api:
            from kubernetes import client

            self._networking_v1_beta1_api = client.NetworkingV1beta1Api(
                api_client=self._k8s_client
            )
        return self._networking_v1_beta1_api



from typing import Union, List, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from ..document import DocumentArray


class GetAttributeMixin:
    """Helpers that provide attributes getter in bulk """

    def get_attributes(self, *fields: str) -> Union[List, List[List]]:
        """Return all nonempty values of the fields from all docs this array contains

        :param fields: Variable length argument with the name of the fields to extract
        :return: Returns a list of the values for these fields.
            When `fields` has multiple values, then it returns a list of list.
        """
        contents = [doc.get_attributes(*fields) for doc in self]

        if len(fields) > 1:
            contents = list(map(list, zip(*contents)))

        return contents

    def get_attributes_with_docs(
        self,
        *fields: str,
    ) -> Tuple[Union[List, List[List]], 'DocumentArray']:
        """Return all nonempty values of the fields together with their nonempty docs

        :param fields: Variable length argument with the name of the fields to extract
        :return: Returns a tuple. The first element is  a list of the values for these fields.
            When `fields` has multiple values, then it returns a list of list. The second element is the non-empty docs.
        """

        contents = []
        docs_pts = []

        for doc in self:
            contents.append(doc.get_attributes(*fields))
            docs_pts.append(doc)

        if len(fields) > 1:
            contents = list(map(list, zip(*contents)))

        from ..document import DocumentArray

        return contents, DocumentArray(docs_pts)



from typing import Dict, Tuple

from google.protobuf.json_format import MessageToJson, MessageToDict

from ..helper import typename, T
from ..proto import jina_pb2


class ProtoTypeMixin:
    """Mixin class of `ProtoType`."""

    def json(self) -> str:
        """Return the object in JSON string

        :return: JSON string of the object
        """
        return MessageToJson(
            self._pb_body, preserving_proto_field_name=True, sort_keys=True
        )

    def dict(self) -> Dict:
        """Return the object in Python dictionary

        :return: dict representation of the object
        """

        # NOTE: PLEASE DO NOT ADD `including_default_value_fields`,
        # it makes the output very verbose!
        return MessageToDict(
            self._pb_body,
            preserving_proto_field_name=True,
        )

    @property
    def proto(self) -> 'jina_pb2._reflection.GeneratedProtocolMessageType':
        """Return the underlying Protobuf object

        :return: Protobuf representation of the object
        """
        return self._pb_body

    def binary_str(self) -> bytes:
        """Return the serialized the message to a string.

        :return: binary string representation of the object
        """
        return self._pb_body.SerializePartialToString()

    @property
    def nbytes(self) -> int:
        """Return total bytes consumed by protobuf.

        :return: number of bytes
        """
        return len(self.binary_str())

    def __getattr__(self, name: str):
        return getattr(self._pb_body, name)

    def __repr__(self):
        content = str(self.non_empty_fields)
        content += f' at {id(self)}'
        return f'<{typename(self)} {content.strip()}>'

    @property
    def non_empty_fields(self) -> Tuple[str, ...]:
        """Return the set fields of the current Protobuf message that are not empty

        :return: the tuple of non-empty fields
        """
        return tuple(field[0].name for field in self._pb_body.ListFields())

    def MergeFrom(self: T, other: T) -> None:
        """Merge the content of target

        :param other: the document to merge from
        """
        self._pb_body.MergeFrom(other._pb_body)

    def CopyFrom(self: T, other: T) -> None:
        """Copy the content of target

        :param other: the document to copy from
        """
        self._pb_body.MergeFrom(other._pb_body)

    def clear(self) -> None:
        """Remove all values from all fields of this Document."""
        self._pb_body.Clear()

    def pop(self, *fields) -> None:
        """Remove the values from the given fields of this Document.

        :param fields: field names
        """
        for k in fields:
            self._pb_body.ClearField(k)

    def __eq__(self, other):
        return self.proto == other.proto



from typing import TYPE_CHECKING

from ....helper import typename

if TYPE_CHECKING:
    from ..types import DocumentArraySourceType


class MagicMixin:
    """Magic helpers for DA/DAM. """

    def __bool__(self):
        """To simulate ```l = []; if l: ...```

        :return: returns true if the length of the array is larger than 0
        """
        return len(self) > 0

    def __repr__(self):
        return f'<{typename(self)} (length={len(self)}) at {id(self)}>'

    def __add__(self, other: 'DocumentArraySourceType'):
        v = type(self)()
        for doc in self:
            v.append(doc)
        for doc in other:
            v.append(doc)
        return v

    def __iadd__(self, other: 'DocumentArraySourceType'):
        for doc in other:
            self.append(doc)
        return self



import warnings
from typing import Optional, Union, Callable, Tuple, Sequence, TYPE_CHECKING

import numpy as np

from .... import Document
from ....math.helper import top_k, minmax_normalize, update_rows_x_mat_best

if TYPE_CHECKING:
    from ..document import DocumentArray
    from ..memmap import DocumentArrayMemmap
    from ...ndarray import ArrayType


class MatchMixin:
    """ A mixin that provides match functionality to DocumentArrays """

    def match(
        self,
        darray: Union['DocumentArray', 'DocumentArrayMemmap'],
        metric: Union[
            str, Callable[['ArrayType', 'ArrayType'], 'np.ndarray']
        ] = 'cosine',
        limit: Optional[Union[int, float]] = 20,
        normalization: Optional[Tuple[float, float]] = None,
        metric_name: Optional[str] = None,
        batch_size: Optional[int] = None,
        traversal_ldarray: Optional[Sequence[str]] = None,
        traversal_rdarray: Optional[Sequence[str]] = None,
        exclude_self: bool = False,
        filter_fn: Optional[Callable[['Document'], bool]] = None,
        only_id: bool = False,
        use_scipy: bool = False,
        device: str = 'cpu',
        **kwargs,
    ) -> None:
        """Compute embedding based nearest neighbour in `another` for each Document in `self`,
        and store results in `matches`.
        .. note::
            'cosine', 'euclidean', 'sqeuclidean' are supported natively without extra dependency.
            You can use other distance metric provided by ``scipy``, such as `braycurtis`, `canberra`, `chebyshev`,
            `cityblock`, `correlation`, `cosine`, `dice`, `euclidean`, `hamming`, `jaccard`, `jensenshannon`,
            `kulsinski`, `mahalanobis`, `matching`, `minkowski`, `rogerstanimoto`, `russellrao`, `seuclidean`,
            `sokalmichener`, `sokalsneath`, `sqeuclidean`, `wminkowski`, `yule`.
            To use scipy metric, please set ``use_scipy=True``.
        - To make all matches values in [0, 1], use ``dA.match(dB, normalization=(0, 1))``
        - To invert the distance as score and make all values in range [0, 1],
            use ``dA.match(dB, normalization=(1, 0))``. Note, how ``normalization`` differs from the previous.
        :param darray: the other DocumentArray or DocumentArrayMemmap to match against
        :param metric: the distance metric
        :param limit: the maximum number of matches, when not given defaults to 20.
        :param normalization: a tuple [a, b] to be used with min-max normalization,
                                the min distance will be rescaled to `a`, the max distance will be rescaled to `b`
                                all values will be rescaled into range `[a, b]`.
        :param metric_name: if provided, then match result will be marked with this string.
        :param batch_size: if provided, then `darray` is loaded in chunks of, at most, batch_size elements. This option
                           will be slower but more memory efficient. Specialy indicated if `darray` is a big
                           DocumentArrayMemmap.
        :param traversal_ldarray: DEPRECATED. if set, then matching is applied along the `traversal_path` of the
                left-hand ``DocumentArray``.
        :param traversal_rdarray: DEPRECATED. if set, then matching is applied along the `traversal_path` of the
                right-hand ``DocumentArray``.
        :param filter_fn: DEPRECATED. if set, apply the filter function to filter docs on the right hand side (rhv) to be matched
        :param exclude_self: if set, Documents in ``darray`` with same ``id`` as the left-hand values will not be
                        considered as matches.
        :param only_id: if set, then returning matches will only contain ``id``
        :param use_scipy: if set, use ``scipy`` as the computation backend. Note, ``scipy`` does not support distance
            on sparse matrix.
        :param device: the computational device for ``.match()``, can be either `cpu` or `cuda`.
        :param kwargs: other kwargs.
        """
        if limit is not None:
            if limit <= 0:
                raise ValueError(f'`limit` must be larger than 0, receiving {limit}')
            else:
                limit = int(limit)

        if batch_size is not None:
            if batch_size <= 0:
                raise ValueError(
                    f'`batch_size` must be larger than 0, receiving {batch_size}'
                )
            else:
                batch_size = int(batch_size)

        lhv = self
        rhv = darray

        if traversal_rdarray or traversal_ldarray or filter_fn:
            warnings.warn(
                '''
            `traversal_ldarray` and `traversal_rdarray` will be removed soon. 
            Instead of doing `da.match(..., traveral_ldarray=[...])`, you can achieve the same via 
            `da.traverse_flat(traveral_ldarray=[...]).match(...)`.
            Same goes with `da.match(..., traveral_rdarray=[...])`, you can do it via: 
            `da.match(da2.traverse_flat(traversal_rdarray=[...]))`.             
            '''
            )

        if traversal_ldarray:
            lhv = self.traverse_flat(traversal_ldarray)

            from ..document import DocumentArray

            if not isinstance(lhv, DocumentArray):
                lhv = DocumentArray(lhv)

        if traversal_rdarray or filter_fn:
            rhv = darray.traverse_flat(traversal_rdarray or ['r'], filter_fn=filter_fn)

            from ..document import DocumentArray

            if not isinstance(rhv, DocumentArray):
                rhv = DocumentArray(rhv)

        if not (lhv and rhv):
            return

        if callable(metric):
            cdist = metric
        elif isinstance(metric, str):
            if use_scipy:
                from scipy.spatial.distance import cdist as cdist
            else:
                from ....math.distance import cdist as _cdist

                cdist = lambda *x: _cdist(*x, device=device)
        else:
            raise TypeError(
                f'metric must be either string or a 2-arity function, received: {metric!r}'
            )

        metric_name = metric_name or (metric.__name__ if callable(metric) else metric)
        _limit = len(rhv) if limit is None else (limit + (1 if exclude_self else 0))

        if batch_size:
            dist, idx = lhv._match_online(
                rhv, cdist, _limit, normalization, metric_name, batch_size
            )
        else:
            dist, idx = lhv._match(rhv, cdist, _limit, normalization, metric_name)

        def _get_id_from_da(rhv, int_offset):
            return rhv._pb_body[int_offset].id

        def _get_id_from_dam(rhv, int_offset):
            return rhv._int2str_id(int_offset)

        from ..memmap import DocumentArrayMemmap

        if isinstance(rhv, DocumentArrayMemmap):
            _get_id = _get_id_from_dam
        else:
            _get_id = _get_id_from_da

        for _q, _ids, _dists in zip(lhv, idx, dist):
            _q.matches.clear()
            num_matches = 0
            for _id, _dist in zip(_ids, _dists):
                # Note, when match self with other, or both of them share the same Document
                # we might have recursive matches .
                # checkout https://github.com/jina-ai/jina/issues/3034
                if only_id:
                    d = Document(id=_get_id(rhv, _id))
                else:
                    d = rhv[int(_id)]  # type: Document

                if d.id in lhv:
                    d = Document(d, copy=True)
                    d.pop('matches')
                if not (d.id == _q.id and exclude_self):
                    _q.matches.append(d, scores={metric_name: _dist})
                    num_matches += 1
                    if num_matches >= (limit or _limit):
                        break

    def _match(self, darray, cdist, limit, normalization, metric_name):
        """
        Computes the matches between self and `darray` loading `darray` into main memory.
        :param darray: the other DocumentArray or DocumentArrayMemmap to match against
        :param cdist: the distance metric
        :param limit: the maximum number of matches, when not given
                      all Documents in `darray` are considered as matches
        :param normalization: a tuple [a, b] to be used with min-max normalization,
                                the min distance will be rescaled to `a`, the max distance will be rescaled to `b`
                                all values will be rescaled into range `[a, b]`.
        :param metric_name: if provided, then match result will be marked with this string.
        :return: distances and indices
        """

        x_mat = self.embeddings
        y_mat = darray.embeddings

        dists = cdist(x_mat, y_mat, metric_name)
        dist, idx = top_k(dists, min(limit, len(darray)), descending=False)
        if isinstance(normalization, (tuple, list)) and normalization is not None:
            # normalization bound uses original distance not the top-k trimmed distance
            min_d = np.min(dists, axis=-1, keepdims=True)
            max_d = np.max(dists, axis=-1, keepdims=True)
            dist = minmax_normalize(dist, normalization, (min_d, max_d))

        return dist, idx

    def _match_online(
        self, darray, cdist, limit, normalization, metric_name, batch_size
    ):
        """
        Computes the matches between self and `darray` loading `darray` into main memory in chunks of size `batch_size`.

        :param darray: the other DocumentArray or DocumentArrayMemmap to match against
        :param cdist: the distance metric
        :param limit: the maximum number of matches, when not given
                      all Documents in `another` are considered as matches
        :param normalization: a tuple [a, b] to be used with min-max normalization,
                              the min distance will be rescaled to `a`, the max distance will be rescaled to `b`
                              all values will be rescaled into range `[a, b]`.
        :param batch_size: length of the chunks loaded into memory from darray.
        :param metric_name: if provided, then match result will be marked with this string.
        :return: distances and indices
        """

        x_mat = self.embeddings
        n_x = x_mat.shape[0]

        idx = 0
        top_dists = np.inf * np.ones((n_x, limit))
        top_inds = np.zeros((n_x, limit), dtype=int)
        for ld in darray.batch(batch_size=batch_size):
            y_batch = ld.embeddings

            distances = cdist(x_mat, y_batch, metric_name)
            dists, inds = top_k(distances, limit, descending=False)

            if isinstance(normalization, (tuple, list)) and normalization is not None:
                dists = minmax_normalize(dists, normalization)

            inds = idx + inds
            idx += y_batch.shape[0]
            top_dists, top_inds = update_rows_x_mat_best(
                top_dists, top_inds, dists, inds, limit
            )

        # sort final the final `top_dists` and `top_inds` per row
        permutation = np.argsort(top_dists, axis=1)
        dist = np.take_along_axis(top_dists, permutation, axis=1)
        idx = np.take_along_axis(top_inds, permutation, axis=1)

        return dist, idx



from contextlib import nullcontext, AsyncExitStack
from typing import Optional, TYPE_CHECKING

from .helper import HTTPClientlet
from ..base import BaseClient
from ..helper import callback_exec
from ...excepts import BadClient
from ...importer import ImportExtensions
from ...logging.profile import ProgressBar
from ...peapods.stream.client import HTTPClientStreamer
from ...types.request import Request

if TYPE_CHECKING:
    from ..base import InputType, CallbackFnType


class HTTPBaseClient(BaseClient):
    """A MixIn for HTTP Client."""

    async def _get_results(
        self,
        inputs: 'InputType',
        on_done: 'CallbackFnType',
        on_error: Optional['CallbackFnType'] = None,
        on_always: Optional['CallbackFnType'] = None,
        **kwargs,
    ):
        """
        :param inputs: the callable
        :param on_done: the callback for on_done
        :param on_error: the callback for on_error
        :param on_always: the callback for on_always
        :param kwargs: kwargs for _get_task_name and _get_requests
        :yields: generator over results
        """
        with ImportExtensions(required=True):
            import aiohttp

        self.inputs = inputs
        request_iterator = self._get_requests(**kwargs)

        async with AsyncExitStack() as stack:
            try:
                cm1 = (
                    ProgressBar(total_length=self._inputs_length)
                    if self.show_progress
                    else nullcontext()
                )
                p_bar = stack.enter_context(cm1)

                proto = 'https' if self.args.https else 'http'
                url = f'{proto}://{self.args.host}:{self.args.port}/post'
                iolet = await stack.enter_async_context(
                    HTTPClientlet(url=url, logger=self.logger)
                )

                streamer = HTTPClientStreamer(self.args, iolet=iolet)
                async for response in streamer.stream(request_iterator):
                    r_status = response.status
                    r_str = await response.json()
                    if r_status == 404:
                        raise BadClient(f'no such endpoint {url}')
                    elif r_status < 200 or r_status > 300:
                        raise ValueError(r_str)

                    resp = Request(r_str)
                    resp = resp.as_typed_request(resp.request_type).as_response()
                    callback_exec(
                        response=resp,
                        on_error=on_error,
                        on_done=on_done,
                        on_always=on_always,
                        continue_on_error=self.continue_on_error,
                        logger=self.logger,
                    )
                    if self.show_progress:
                        p_bar.update()
                    yield resp
            except aiohttp.ClientError as e:
                self.logger.error(
                    f'Error while fetching response from HTTP server {e!r}'
                )



import os
import tempfile
import json
from typing import Dict, Optional, Generator

from .kubernetes_client import K8sClients
from ....importer import ImportExtensions
from ....logging.logger import JinaLogger
from ....logging.predefined import default_logger

cur_dir = os.path.dirname(__file__)
DEFAULT_RESOURCE_DIR = os.path.join(
    cur_dir, '..', '..', '..', 'resources', 'k8s', 'template'
)


def create(
    template: str,
    params: Dict,
    logger: JinaLogger = default_logger,
    custom_resource_dir: Optional[str] = None,
):
    """Create a resource on Kubernetes based on the `template`. It fills the `template` using the `params`.

    :param template: path to the template file.
    :param custom_resource_dir: Path to a folder containing the kubernetes yml template files.
        Defaults to the standard location jina.resources if not specified.
    :param logger: logger to use. Defaults to the default logger.
    :param params: dictionary for replacing the placeholders (keys) with the actual values.
    """

    from kubernetes.utils import FailToCreateError
    from kubernetes import utils

    clients = K8sClients()
    if template == 'configmap':
        yaml = _patch_configmap_yaml(template, params)
    elif template in ['deployment', 'deployment-init'] and params.get('device_plugins'):
        yaml = _get_yaml(template, params, custom_resource_dir)
        yaml = _patch_deployment_with_device_plugins(yaml, params)
    else:
        yaml = _get_yaml(template, params, custom_resource_dir)
    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, 'w') as tmp:
            tmp.write(yaml)
        try:
            utils.create_from_yaml(clients.k8s_client, path)
        except FailToCreateError as e:
            for api_exception in e.api_exceptions:
                if api_exception.status == 409:
                    # The exception's body is the error response from the
                    # Kubernetes apiserver, it looks like:
                    # {..."message": "<resource> <name> already exists"...}
                    resp = json.loads(api_exception.body)
                    logger.warning(f'🔁\t{resp["message"]}')
                else:
                    raise e
        except Exception as e2:
            raise e2
    finally:
        os.remove(path)


def replace(
    deployment_name: str,
    namespace_name: str,
    template: str,
    params: Dict,
    custom_resource_dir: Optional[str] = None,
):
    """Create a resource on Kubernetes based on the `template`. It fills the `template` using the `params`.

    :param deployment_name: The name of the deployment to replace
    :param namespace_name: The name of the namespace where the deployment exists
    :param template: path to the template file.
    :param custom_resource_dir: Path to a folder containing the kubernetes yml template files.
        Defaults to the standard location jina.resources if not specified.
    :param params: dictionary for replacing the placeholders (keys) with the actual values.
    """

    import yaml

    yaml_file_path = _get_yaml(template, params, custom_resource_dir)
    fd, path = tempfile.mkstemp()
    try:
        with os.fdopen(fd, 'w') as tmp:
            tmp.write(yaml_file_path)
        with open(os.path.abspath(path)) as f:
            yml_document_all = yaml.safe_load(f)
        clients = K8sClients()
        clients.apps_v1.replace_namespaced_deployment(
            deployment_name, namespace_name, yml_document_all
        )
    finally:
        os.remove(path)


def _get_yaml(template: str, params: Dict, custom_resource_dir: Optional[str] = None):
    if custom_resource_dir:
        path = os.path.join(custom_resource_dir, f'{template}.yml')
    else:
        path = os.path.join(DEFAULT_RESOURCE_DIR, f'{template}.yml')
    with open(path) as f:
        content = f.read()
        for k, v in params.items():
            content = content.replace(f'{{{k}}}', str(v))
    return content


def _patch_configmap_yaml(template: str, params: Dict):
    import yaml

    path = os.path.join(DEFAULT_RESOURCE_DIR, f'{template}.yml')

    with open(path) as f:
        config_map = yaml.safe_load(f)

    config_map['metadata']['name'] = params.get('name') + '-' + 'configmap'
    config_map['metadata']['namespace'] = params.get('namespace')
    if params.get('data'):
        for key, value in params['data'].items():
            config_map['data'][key] = value
    return json.dumps(config_map)


def _create_device_plugins(params: Dict):
    data = {'limits': {}}
    for key, value in params.items():
        data['limits'][key] = value
    return data


def _patch_deployment_with_device_plugins(yaml_content: str, params: Dict):
    import yaml

    device_plugins = _create_device_plugins(params['device_plugins'])

    deployment = yaml.safe_load(yaml_content)
    deployment['spec']['template']['spec']['containers'][0][
        'resources'
    ] = device_plugins
    return json.dumps(deployment)


def _get_gateway_pod_name(namespace, k8s_clients):
    gateway_pod = k8s_clients.core_v1.list_namespaced_pod(
        namespace=namespace, label_selector='app=gateway'
    )
    return gateway_pod.items[0].metadata.name


def get_port_forward_contextmanager(
    namespace: str,
    port_expose: int,
    config_path: str = None,
) -> Generator[None, None, None]:
    """Forward local requests to the gateway which is running in the Kubernetes cluster.
    :param namespace: namespace of the gateway
    :param port_expose: exposed port of the gateway
    :param config_path: path to the Kubernetes config file
    :return: context manager which sets up and terminates the port-forward
    """
    with ImportExtensions(
        required=True,
        help_text='Sending requests to the Kubernetes cluster requires to install the portforward package. '
        'Please do `pip install "jina[portforward]"`'
        'Also make sure golang is installed `https://golang.org/`',
    ):
        import portforward

    clients = K8sClients()
    gateway_pod_name = _get_gateway_pod_name(namespace, k8s_clients=clients)
    if config_path is None and 'KUBECONFIG' in os.environ:
        config_path = os.environ['KUBECONFIG']
    return portforward.forward(
        namespace, gateway_pod_name, port_expose, port_expose, config_path
    )



from typing import Union, BinaryIO

from .helper import _uri_to_buffer, _get_file_context
from ....helper import T


class DumpFileMixin:
    """Provide helper functions for :class:`Document` to dump content to a file. """

    def dump_buffer_to_file(self: T, file: Union[str, BinaryIO]) -> T:
        """Save :attr:`.buffer` into a file

        :param file: File or filename to which the data is saved.
        :return: itself after processed
        """
        fp = _get_file_context(file)
        with fp:
            fp.write(self.buffer)
        return self

    def dump_uri_to_file(self: T, file: Union[str, BinaryIO]) -> T:
        """Save :attr:`.uri` into a file

        :param file: File or filename to which the data is saved.

        :return: itself after processed
        """
        fp = _get_file_context(file)
        with fp:
            buffer = _uri_to_buffer(self.uri)
            fp.write(buffer)
        return self



from typing import TYPE_CHECKING, Union

from .helper import _deprecate_by
from ...ndarray import get_array_type
from ....helper import T
from ....helper import typename
from ....proto import jina_pb2

if TYPE_CHECKING:
    from ...ndarray import ArrayType

    DocumentContentType = Union[bytes, str, ArrayType]

_DIGEST_SIZE = 8


class ContentPropertyMixin:
    """Provide helper functions for :class:`Document` to allow universal content property access. """

    @property
    def content(self) -> 'DocumentContentType':
        """Return the content of the document. It checks whichever field among :attr:`blob`, :attr:`text`,
        :attr:`buffer` has value and return it.

        .. seealso::
            :attr:`blob`, :attr:`buffer`, :attr:`text`

        :return: the value of the content depending on `:meth:`content_type`
        """
        attr = self.content_type
        if attr:
            return getattr(self, attr)

    @content.setter
    def content(self, value: 'DocumentContentType'):
        """Set the content of the document. It assigns the value to field with the right type.

        .. seealso::
            :attr:`blob`, :attr:`buffer`, :attr:`text`

        :param value: the value from which to set the content of the Document
        """
        if isinstance(value, bytes):
            self.buffer = value
        elif isinstance(value, str):
            self.text = value
        else:
            try:
                get_array_type(value)
                self.blob = value
            except:
                raise TypeError(f'{typename(value)} is not recognizable')

    @property
    def content_type(self) -> str:
        """Return the content type of the document, possible values: text, blob, buffer

        :return: the type of content of this Document
        """
        return self._pb_body.WhichOneof('content')

    @property
    def content_hash(self) -> str:
        """Get the document hash according to its content.

        :return: the unique hash code to represent this Document
        """
        # a tuple of field names that inclusive when computing content hash.
        from google.protobuf.field_mask_pb2 import FieldMask
        from hashlib import blake2b

        fields = (
            'text',
            'blob',
            'buffer',
            'embedding',
            'uri',
            'tags',
            'mime_type',
            'granularity',
            'adjacency',
        )
        masked_d = jina_pb2.DocumentProto()
        present_fields = {
            field_descriptor.name for field_descriptor, _ in self._pb_body.ListFields()
        }
        fields_to_hash = present_fields.intersection(fields)
        FieldMask(paths=fields_to_hash).MergeMessage(self._pb_body, masked_d)

        return blake2b(
            masked_d.SerializePartialToString(), digest_size=_DIGEST_SIZE
        ).hexdigest()

    def dump_content_to_datauri(self: T) -> T:
        """Convert :attr:`.content` in :attr:`.uri` inplace with best effort

        :return: itself after processed
        """
        if self.text:
            self.convert_text_to_uri()
        elif self.buffer:
            self.convert_buffer_to_uri()
        elif self.content_type:
            raise NotImplementedError
        return self

    convert_content_to_uri = _deprecate_by(dump_content_to_datauri)



import warnings
from abc import abstractmethod
from typing import (
    Iterable,
    Sequence,
    TYPE_CHECKING,
    Optional,
    Callable,
    Union,
)

if TYPE_CHECKING:
    from ..document import DocumentArray
    from ...document import Document
    from ....helper import T


def _check_traversal_path_type(tp):
    is_deprecate_scheme = True
    if isinstance(tp, str):
        tp = tp.split(',')
        is_deprecate_scheme = False

    if not (
        tp
        and not isinstance(tp, str)
        and isinstance(tp, Sequence)
        and all(isinstance(p, str) for p in tp)
    ):
        raise TypeError('`traversal_paths` needs to be `Sequence[str]`')

    if is_deprecate_scheme:
        warnings.warn(
            f'The syntax of traversal_path is changed to comma-separated string, '
            f'that means your need to change {tp} into `{",".join(tp)}`. '
            f'The old list of string syntax will be deprecated soon',
            DeprecationWarning,
        )

    return tp


class TraverseMixin:
    """
    A mixin used for traversing :class:`DocumentArray` or :class:`DocumentArrayMemmap`.
    """

    def traverse(
        self: 'T',
        traversal_paths: str,
        filter_fn: Optional[Callable[['Document'], bool]] = None,
    ) -> Iterable['T']:
        """
        Return an Iterator of :class:``TraversableSequence`` of the leaves when applying the traversal_paths.
        Each :class:``TraversableSequence`` is either the root Documents, a ChunkArray or a MatchArray.

        :param traversal_paths: a comma-separated string that represents the traversal path
        :param filter_fn: function to filter docs during traversal
        :yield: :class:``TraversableSequence`` of the leaves when applying the traversal_paths.

        Example on ``traversal_paths``:

            - `r`: docs in this TraversableSequence
            - `m`: all match-documents at adjacency 1
            - `c`: all child-documents at granularity 1
            - `cc`: all child-documents at granularity 2
            - `mm`: all match-documents at adjacency 2
            - `cm`: all match-document at adjacency 1 and granularity 1
            - `r,c`: docs in this TraversableSequence and all child-documents at granularity 1

        """
        traversal_paths = _check_traversal_path_type(traversal_paths)

        for p in traversal_paths:
            yield from self._traverse(self, p, filter_fn=filter_fn)

    @staticmethod
    def _traverse(
        docs: 'T',
        path: str,
        filter_fn: Optional[Callable[['Document'], bool]] = None,
    ):
        if path:
            loc = path[0]
            if loc == 'r':
                yield from TraverseMixin._traverse(docs, path[1:], filter_fn=filter_fn)
            elif loc == 'm':
                for d in docs:
                    yield from TraverseMixin._traverse(
                        d.matches, path[1:], filter_fn=filter_fn
                    )
            elif loc == 'c':
                for d in docs:
                    yield from TraverseMixin._traverse(
                        d.chunks, path[1:], filter_fn=filter_fn
                    )
            else:
                raise ValueError(
                    f'`path`:{loc} is invalid, must be one of `c`, `r`, `m`'
                )
        elif filter_fn is None:
            yield docs
        else:
            from ..document import DocumentArray

            yield DocumentArray(list(filter(filter_fn, docs)))

    def traverse_flat_per_path(
        self,
        traversal_paths: str,
        filter_fn: Optional[Callable[['Document'], bool]] = None,
    ):
        """
        Returns a flattened :class:``TraversableSequence`` per path in ``traversal_paths``
        with all Documents, that are reached by the path.

        :param traversal_paths: a comma-separated string that represents the traversal path
        :param filter_fn: function to filter docs during traversal
        :yield: :class:``TraversableSequence`` containing the document of all leaves per path.
        """
        traversal_paths = _check_traversal_path_type(traversal_paths)

        for p in traversal_paths:
            yield self._flatten(self._traverse(self, p, filter_fn=filter_fn))

    def traverse_flat(
        self,
        traversal_paths: str,
        filter_fn: Optional[Callable[['Document'], bool]] = None,
    ) -> Union['DocumentArray', Iterable['Document']]:
        """
        Returns a single flattened :class:``TraversableSequence`` with all Documents, that are reached
        via the ``traversal_paths``.

        .. warning::
            When defining the ``traversal_paths`` with multiple paths, the returned
            :class:``Documents`` are determined at once and not on the fly. This is a different
            behavior then in :method:``traverse`` and :method:``traverse_flattened_per_path``!

        :param traversal_paths: a list of string that represents the traversal path
        :param filter_fn: function to filter docs during traversal
        :return: a single :class:``TraversableSequence`` containing the document of all leaves when applying the traversal_paths.
        """
        traversal_paths = _check_traversal_path_type(traversal_paths)
        if (
            len(traversal_paths) == 1
            and traversal_paths[0] == 'r'
            and filter_fn is None
        ):
            return self

        leaves = self.traverse(traversal_paths, filter_fn=filter_fn)
        return self._flatten(leaves)

    def flatten(self, copy: bool = True) -> 'DocumentArray':
        """Flatten all nested chunks and matches into one :class:`DocumentArray`.

        .. note::
            Flatten an already flattened DocumentArray will have no effect.

        .. warning::
            DocumentArrayMemmap do not support `copy=False`.

        :param copy: copy the document (DAM only supports copy=True), otherwise returns a view of the original
        :return: a flattened :class:`DocumentArray` object.
        """
        from ..document import DocumentArray

        def _yield_all():
            for d in self:
                yield from _yield_nest(d)

        def _yield_nest(doc: 'Document'):
            from ...document import Document

            for d in doc.chunks:
                yield from _yield_nest(d)
            for m in doc.matches:
                yield from _yield_nest(m)

            if copy:
                d = Document(doc, copy=True)
                d.chunks.clear()
                d.matches.clear()
                yield d
            else:
                doc.matches.clear()
                doc.chunks.clear()
                yield doc

        return DocumentArray(_yield_all())

    @staticmethod
    @abstractmethod
    def _flatten(sequence):
        ...
